# -*- coding: utf-8 -*-
#
#  Copyright 2019-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2019 Tagir Akhmetshin <tagirshin@gmail.com>
#  Copyright 2019 Tansu Nasyrova <tansu.nasyrova@gmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from .element import Element
from .groups import GroupXIII
from .periods import PeriodII, PeriodIII, PeriodIV, PeriodV, PeriodVI, PeriodVII


class B(Element, PeriodII, GroupXIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 5

    @property
    def isotopes_distribution(self):
        return {10: 0.199, 11: 0.801}

    @property
    def isotopes_masses(self):
        return {10: 10.012937, 11: 11.009305}

    @property
    def _common_valences(self):
        return 3,

    @property
    def _valences_exceptions(self):
        return (-1, False, 4, ()),

    @property
    def atomic_radius(self):
        return .87


class Al(Element, PeriodIII, GroupXIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 13

    @property
    def isotopes_distribution(self):
        return {27: 1.0}

    @property
    def isotopes_masses(self):
        return {27: 26.981538}

    @property
    def _common_valences(self):
        return 0,  # 3 omitted. added to exceptions

    @property
    def _valences_exceptions(self):
        return ((3, False, 0, ()), (2, False, 1, ()), (1, False, 2, ()), (0, False, 3, ()), (-1, False, 4, ()),
                (-3, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))))

    @property
    def atomic_radius(self):
        return 1.18


class Ga(Element, PeriodIV, GroupXIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 31

    @property
    def isotopes_distribution(self):
        return {67: 0., 69: 0.60108, 71: 0.39892}

    @property
    def isotopes_masses(self):
        return {67: 66.928202, 69: 68.925581, 71: 70.924705}

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return ((3, False, 0, ()),
                (0, False, 0, ((1, 'Cl'),)), (0, False, 0, ((1, 'Br'),)), (0, False, 0, ((1, 'I'),)),
                (-1, False, 0, ((1, 'H'), (1, 'H'), (1, 'H'), (1, 'H'))),
                (-1, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))),
                (-1, False, 0, ((1, 'Cl'), (1, 'Cl'), (1, 'Cl'), (1, 'Cl'))),
                (-1, False, 0, ((1, 'Br'), (1, 'Br'), (1, 'Br'), (1, 'Br'))),
                (-1, False, 0, ((1, 'I'), (1, 'I'), (1, 'I'), (1, 'I'))))

    @property
    def atomic_radius(self):
        return 1.36


class In(Element, PeriodV, GroupXIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 49

    @property
    def isotopes_distribution(self):
        return {111: 0., 113: 0.0429, 115: 0.9571}

    @property
    def isotopes_masses(self):
        return {111: 110.905103, 113: 112.904061, 115: 114.903878}

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return ((3, False, 0, ()),
                (0, False, 0, ((1, 'Cl'),)), (0, False, 0, ((1, 'Br'),)), (0, False, 0, ((1, 'I'),)),
                (0, False, 0, ((1, 'O'),)))

    @property
    def atomic_radius(self):
        return 1.56


class Tl(Element, PeriodVI, GroupXIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 81

    @property
    def isotopes_distribution(self):
        return {203: 0.29524, 205: 0.70476}

    @property
    def isotopes_masses(self):
        return {203: 202.972329, 205: 204.974412}

    @property
    def _common_valences(self):
        return 0, 1

    @property
    def _valences_exceptions(self):
        return ((1, False, 0, ()), (3, False, 0, ()),
                (-3, False, 0, ((1, 'Cl'), (1, 'Cl'), (1, 'Cl'), (1, 'Cl'), (1, 'Cl'), (1, 'Cl'))),
                (1, False, 0, ((1, 'C'), (1, 'C'))))

    @property
    def atomic_radius(self):
        return 1.56


class Nh(Element, PeriodVII, GroupXIII):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 113

    @property
    def isotopes_distribution(self):
        return {286: 1.0}

    @property
    def isotopes_masses(self):
        return {286: 286.182555}

    @property
    def _common_valences(self):
        return 0,

    @property
    def _valences_exceptions(self):
        return ()

    @property
    def atomic_radius(self):
        return 1.56  # unknown, taken radius of previous element in group


__all__ = ['B', 'Al', 'Ga', 'In', 'Tl', 'Nh']



# -*- coding: utf-8 -*-
#
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from chython import SMILESRead, smiles, MoleculeContainer, CGRContainer
from chython.files._mdl import parse_error
from itertools import zip_longest
from io import StringIO


def test_bad_smiles_detection():
    data = ('[[C] C%[C] [C]] [] [(C)] C%(C) C(C%) C() C(C)) C((C) C%01 C01 S%% CBl CCr Cd C[H C(C [Zn=>=] [12] [+]' 
            '[C:1-] [C+5>0] [1+>-] [=*] [C+>-5] (=C) 1CC1 C#(C) C(C#) C=-C C.1C1 C1CC1C1CC1 C=1CC1 C#1CC-1 C1CC-1'
            'C(C)C( C1C CC= [->=]').replace(' ', '\n')

    with StringIO(data) as f, SMILESRead(f, ignore=False, store_log=True) as r:
        for s in r._data:
            assert isinstance(s, parse_error)


def test_invalid_ring_closure_ignoring():
    data = 'C1CC1C1CC1\nC=1CC1'
    with StringIO(data) as f, SMILESRead(f, ignore=True, store_log=True) as r:
        for t, v in zip_longest(r, ('C1CC1C2CC2', 'C1=CC1')):
            assert t == smiles(v)


def test_good_smiles():
    for t, v in zip('CN C-N C=N C#N C:N cn'.split(), (1, 1, 2, 3, 4, 4)):
        t = smiles(t)
        assert isinstance(t, MoleculeContainer)
        assert t.bond(1, 2).order == v

    t = smiles('C.N ')
    assert isinstance(t, MoleculeContainer)
    assert not t.has_bond(1, 2)

    for t, v1, v2 in zip('C[->.]N C[->=]N C[=>#]N C[#>:]N C[:>.]N [O->0][.>-][Na+>0]'.split(),
                         (1, 1, 2, 3, 4, None), (None, 2, 3, 4, None, 1)):
        t = smiles(t)
        assert isinstance(t, CGRContainer)
        assert t.bond(1, 2).order == v1
        assert t.bond(1, 2).p_order == v2

    for t in '[Fe] [13C] [C+4] [C:14] [C--:12] [C@@H] [C@H] [OH2] [OH3+]'.split():
        assert isinstance(smiles(t), MoleculeContainer)

    for t in 'C(C)C C(-C)C C(C)-C'.split():
        assert smiles(t).bonds_count == 2

    for t, v in zip('C1CC1 C-1CC-1 C%10CC%10 C-%11CC-%11 C12CC1C2 C-1-2CC-1C-2'.split(), (1, 1, 1, 1, 2, 2)):
        assert smiles(t).rings_count == v

    for t in ('[C+>-] [C++>--] [C-3>+2] [C0>+] [C-->0] [C*>^] [C^>*] [C+>-*>^] [C-3>+2^>*] '
              '[C0>-*>^] [C+2>0^>*] [n+>0]').split():
        assert isinstance(smiles(t), CGRContainer)



# -*- coding: utf-8 -*-
#
#  Copyright 2020, 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from .parser import Parser
from ...exceptions import NotChiral, IsChiral, ValenceError


class MDLStereo(Parser):
    def __init__(self, calc_cis_trans=False, ignore_stereo=False, **kwargs):
        super().__init__(**kwargs)
        self.__calc_cis_trans = calc_cis_trans
        self.__ignore_stereo = ignore_stereo

    def _create_molecule(self, data, mapping):
        mol = super()._create_molecule(data, mapping)
        hydrogens = mol._hydrogens
        for n, h in data['hydrogens'].items():
            n = mapping[n]
            hc = hydrogens[n]
            if hc is None:  # aromatic rings or valence errors. just store given H count.
                hydrogens[n] = h
            elif hc != h:  # H count mismatch. try radical state of atom.
                if self._ignore:
                    hydrogens[n] = h  # set parsed hydrogens count
                    self._info(f'implicit hydrogen count ({h}) mismatch with '
                               f'calculated ({hc}) on atom {n}. calculated count replaced.')
                else:
                    raise ValueError(f'implicit hydrogen count ({h}) mismatch with '
                                     f'calculated ({hc}) on atom {n}.')

        if self.__ignore_stereo:
            return mol

        if self.__calc_cis_trans:
            mol.calculate_cis_trans_from_2d()

        stereo = [(mapping[n], mapping[m], s) for n, m, s in data['stereo']]
        while stereo:
            fail_stereo = []
            old_stereo = len(stereo)
            for n, m, s in stereo:
                try:
                    mol.add_wedge(n, m, s, clean_cache=False)
                except NotChiral:
                    fail_stereo.append((n, m, s))
                except IsChiral:
                    pass
                except ValenceError:
                    self._info('structure has errors, stereo data skipped')
                    mol.flush_cache()
                    break
            else:
                stereo = fail_stereo
                if len(stereo) == old_stereo:
                    break
                mol.flush_stereo_cache()
                if self.__calc_cis_trans:
                    mol.calculate_cis_trans_from_2d(clean_cache=False)
                continue
            break
        return mol


__all__ = ['MDLStereo']



# -*- coding: utf-8 -*-
#
#  Copyright 2020, 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from functools import cached_property
from numpy import zeros
from numpy.linalg import eig
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from chython import MoleculeContainer


# atom, charge, radical, non sp3 : h, k, ne
basis = {(5, 0, False, False): (-1., .75, 0),  # X3B
         (6, 0, False, True): (0., 1., 1),
         (7, 0, False, True): (.5, 1., 1),  # pyridine
         (7, 0, False, False): (1.49, .8, 2),  # pyrole
         (7, 1, False, True): (2.3, .7, 1),  # nitro
         (8, 0, False, True): (1., 1., 1),  # X=O
         (8, 0, False, False): (2., .8, 2),  # X-O-Y
         (8, -1, False, False): (2., .8, 2),  # X-[O-]
         (16, 0, False, True): (.4, 1., 1),  # X=S
         (16, 0, False, False): (1.3, .6, 2),  # X-S-Y
         (16, -1, False, False): (1.3, .6, 2),  # X-[S-]
         }


class Huckel:
    __slots__ = ()

    @cached_property
    def huckel_energy(self: 'MoleculeContainer') -> float:
        """
        Huckel method based Pi electrons energy calculator.
        Parametrized for B C N O S.
        """
        hyb = self.hybridization
        charge = self._charges
        radical = self._radicals

        adj = {}
        alpha = {}
        beta = {}
        electrons = {}
        for n, a in self._atoms.items():  # collect Pi atoms
            an = a.atomic_number
            ac = charge[n]
            ar = radical[n]
            ah = hyb(n) > 1
            if an == 6:
                if ah or ac or ar:  # unsaturated carbon
                    adj[n] = {}
                    alpha[n], beta[n], electrons[n] = basis[(6, ac, ar, ah)]
            elif an in (5, 7, 8, 16):
                try:
                    alpha[n], beta[n], electrons[n] = basis[(an, ac, ar, ah)]
                except KeyError:  # not parametrized or don't have Pi orbitals
                    continue
                adj[n] = {}
        if not adj:
            return 0.
        for n, m, _ in self.bonds():
            if n in adj and m in adj:
                adj[n][m] = adj[m][n] = min(beta[n], beta[m])

        energy = 0.
        for comp in self._connected_components(adj):
            h_matrix = zeros((len(comp), len(comp)))
            mapping = {}
            e = 0
            for i, n in enumerate(comp):
                h_matrix[i, i] = alpha[n]
                mapping[n] = i
                e += electrons[n]
            for n in comp:
                i = mapping[n]
                for m, b in adj[n].items():
                    j = mapping[m]
                    h_matrix[i, j] = h_matrix[j, i] = b

            orbs = sorted(list(eig(h_matrix)[0]))
            paired = e // 2
            energy += sum(x * 2 for x in orbs[:paired])
            if e % 2:  # unpaired
                energy += orbs[paired]
        return energy


__all__ = ['Huckel']



# -*- coding: utf-8 -*-
#
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
cimport cython
from libc.math cimport ldexp
from chython.containers.bonds import Bond


@cython.boundscheck(False)
def unpack(bytes data not None):
    cdef int isotope_shift
    cdef unsigned char a, b, c, d
    cdef unsigned int na, nct, i, j, n, m, shift = 4, order_shift = 0, nb = 0

    cdef unsigned int[4095] mapping, atom, isotopes, hydrogens, neighbors, orders, cis_trans_1, cis_trans_2
    cdef unsigned int[8190] connections
    cdef int[4095] charges
    cdef bint[4095] radicals, is_tet, is_all, tet_sign, all_sign, ct_sign
    cdef double[4095] x, y
    cdef bint[4096] seen

    cdef object bond
    cdef dict py_charges, py_radicals, py_hydrogens, py_plane, py_bonds, tmp
    cdef dict py_atoms_stereo, py_allenes_stereo, py_cis_trans_stereo
    cdef list py_mapping, py_atoms, py_isotopes

    # lets extract data
    a, b, c = data[1], data[2], data[3]
    na = a << 4| (b & 0xf0) >> 4
    nct = (b & 0x0f) << 8 | c

    for i in range(na):
        a, b = data[shift], data[shift + 1]
        mapping[i] = n = a << 4 | (b & 0xf0) >> 4
        seen[n] = False
        neighbors[i] = b & 0x0f
        nb += b & 0x0f

        a, b = data[shift + 2], data[shift + 3]
        if a & 0x80:
            is_tet[i] = True
            tet_sign[i] = a & 0x40
        else:
            is_tet[i] = False
        if a & 0x20:
            is_all[i] = True
            all_sign[i] = a & 0x10
        else:
            is_all[i] = False

        atom[i] = b & 0x7f
        isotope_shift = (a & 0x0f) << 1 | b >> 7
        if isotope_shift:
            isotopes[i] = common_isotopes[b & 0x7f] + isotope_shift
        else:
            isotopes[i] = 0

        a, b = data[shift + 4], data[shift + 5]
        x[i] = double_from2bytes(a, b)
        a, b = data[shift + 6], data[shift + 7]
        y[i] = double_from2bytes(a, b)

        a = data[shift + 8]
        hydrogens[i] = a >> 5
        charges[i] = ((a >> 1) & 0x0f) - 4
        radicals[i] = a & 0x01

        shift += 9

    nb //= 2
    for i in range(nb):
        a, b, c = data[shift], data[shift + 1], data[shift + 2]
        connections[i * 2] = a << 4| (b & 0xf0) >> 4
        connections[i * 2 + 1] = (b & 0x0f) << 8 | c
        shift += 3

    for i in range((nb // 5 + 1) if nb % 5 else (nb // 5)):
        a, b = data[shift], data[shift + 1]
        orders[i * 5] = (a >> 4) + 1
        orders[i * 5 + 1] = ((a >> 1) & 0x07) + 1
        orders[i * 5 + 2] = ((a & 0x01) << 2 | b >> 6) + 1
        orders[i * 5 + 3] = ((b >> 3) & 0x07) + 1
        orders[i * 5 + 4] = (b & 0x07) + 1
        shift += 2

    for i in range(nct):
        a, b, c, d = data[shift], data[shift + 1], data[shift + 2], data[shift + 3]
        cis_trans_1[i] = a << 4 | (b & 0xf0) >> 4
        cis_trans_2[i] = (b & 0x0f) << 8 | c
        ct_sign[i] = d & 0x01
        shift += 4

    # define returned data
    py_mapping = []
    py_atoms = []
    py_isotopes = []
    py_bonds = {}
    py_charges = {}
    py_radicals = {}
    py_hydrogens = {}
    py_plane = {}
    py_atoms_stereo = {}
    py_allenes_stereo = {}
    py_cis_trans_stereo = {}

    shift = 0
    for i in range(na):
        n = mapping[i]

        # fill intermediate data
        py_mapping.append(n)
        py_atoms.append(atom[i])
        py_isotopes.append(isotopes[i] or None)

        py_charges[n] = charges[i]
        py_radicals[n] = radicals[i]
        if hydrogens[i] == 7:
            py_hydrogens[n] = None
        else:
            py_hydrogens[n] = hydrogens[i]
        py_plane[n] = (x[i], y[i])

        if is_tet[i]:
            py_atoms_stereo[n] = tet_sign[i]
        if is_all[i]:
            py_allenes_stereo[n] = all_sign[i]

        tmp = {}
        py_bonds[n] = tmp
        seen[n] = True
        for j in range(shift, shift + neighbors[i]):
            m = connections[j]
            if seen[m]:  # bond partially exists. need back-connection.
                tmp[m] = py_bonds[m][n]
            else:
                bond = object.__new__(Bond)
                bond._Bond__order = orders[order_shift]
                tmp[m] = bond
                order_shift += 1

        shift += neighbors[i]

    for i in range(nct):
        py_cis_trans_stereo[(cis_trans_1[i], cis_trans_2[i])] = ct_sign[i]

    return (py_mapping, py_atoms, py_isotopes,
            py_charges, py_radicals, py_hydrogens, py_plane, py_bonds,
            py_atoms_stereo, py_allenes_stereo, py_cis_trans_stereo)


cdef int[119] common_isotopes
common_isotopes[:] = [0, -15, -12, -9, -7, -5, -4, -2, 0, 3, 4, 7, 8, 11, 12, 15, 16, 19, 24, 23, 24, 29,
                      32, 35, 36, 39, 40, 43, 43, 48, 49, 54, 57, 59, 63, 64, 68, 69, 72, 73, 75, 77,
                      80, 82, 85, 87, 90, 92, 96, 99, 103, 106, 112, 111, 115, 117, 121, 123, 124, 125,
                      128, 129, 134, 136, 141, 143, 147, 149, 151, 153, 157, 159, 162, 165, 168, 170,
                      174, 176, 179, 181, 185, 188, 191, 193, 193, 194, 206, 207, 210, 211, 216, 215,
                      222, 221, 228, 227, 231, 231, 235, 236, 241, 242, 243, 244, 245, 254, 253, 254,
                      254, 262, 265, 265, 269, 262, 273, 273, 277, 281, 278]


cdef double double_from2bytes(unsigned char a, unsigned char b):
    cdef bint sign
    cdef int e
    cdef unsigned int f
    cdef double x

    sign = a >> 7
    e = (a >> 2) & 0x1f
    f = ((a & 0x03) << 8) | b

    x = f / 1024.
    if e:
        x += 1.
        e -= 15
    else:
        e = -14

    x = ldexp(x, e)
    if sign:
        return -x
    return x



# -*- coding: utf-8 -*-
#
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from lazy_object_proxy import Proxy
from typing import List, Tuple


def _fixed_rules():
    """
    Rules working without Morgan.
    This rules are used for charge canonization for heterocycles
    """
    from ...containers import QueryContainer

    rules: List[Tuple[QueryContainer, bool]] = []

    # N1C=CN2[NH+]=CC=C12
    q = QueryContainer()
    q.add_atom('N', charge=1)  # first atom is charged
    q.add_atom('N')  # second should be charged atom
    q.add_atom('N')
    for _ in range(5):
        q.add_atom('A')
    q.add_bond(1, 3, 4)
    q.add_bond(3, 4, 4)
    q.add_bond(3, 6, 4)
    q.add_bond(4, 5, 4)
    q.add_bond(5, 2, 4)
    q.add_bond(2, 6, 4)
    q.add_bond(6, 7, 4)
    q.add_bond(7, 8, 4)
    q.add_bond(8, 1, 4)
    rules.append((q, False))

    # N1C=C[N+]2=C1C=CN2
    q = QueryContainer()
    q.add_atom('N')
    q.add_atom('N')  # second should be charged atom
    q.add_atom('N', charge=1)  # bad charged
    for _ in range(5):
        q.add_atom('A')
    q.add_bond(1, 3, 4)
    q.add_bond(3, 4, 4)
    q.add_bond(3, 6, 4)
    q.add_bond(4, 5, 4)
    q.add_bond(5, 2, 4)
    q.add_bond(2, 6, 4)
    q.add_bond(6, 7, 4)
    q.add_bond(7, 8, 4)
    q.add_bond(8, 1, 4)
    rules.append((q, True))

    # N1C=C2C=CN[N+]2=C1
    q = QueryContainer()
    q.add_atom('N')
    q.add_atom('N')  # second should be charged atom
    q.add_atom('N', charge=1)
    for _ in range(5):
        q.add_atom('A')
    q.add_bond(1, 3, 4)
    q.add_bond(3, 4, 4)
    q.add_bond(3, 6, 4)
    q.add_bond(4, 2, 4)
    q.add_bond(2, 5, 4)
    q.add_bond(5, 6, 4)
    q.add_bond(6, 7, 4)
    q.add_bond(7, 8, 4)
    q.add_bond(8, 1, 4)
    rules.append((q, True))

    # N1C=C2NC=C[N+]2=C1
    q = QueryContainer()
    q.add_atom('N')
    q.add_atom('N')  # second should be charged atom
    q.add_atom('N', charge=1)
    for _ in range(5):
        q.add_atom('A')
    q.add_bond(1, 6, 4)
    q.add_bond(3, 4, 4)
    q.add_bond(3, 6, 4)
    q.add_bond(4, 2, 4)
    q.add_bond(2, 5, 4)
    q.add_bond(5, 6, 4)
    q.add_bond(3, 7, 4)
    q.add_bond(7, 8, 4)
    q.add_bond(8, 1, 4)
    rules.append((q, True))

    # C1=CC=2C(=[NH+]1)C=CNC=2
    q = QueryContainer()
    q.add_atom('N', charge=1)  # first atom is charged
    q.add_atom('N')  # second is possible charged atom
    for _ in range(7):
        q.add_atom('A')
    q.add_bond(2, 9, 4)
    q.add_bond(9, 3, 4)
    q.add_bond(3, 4, 4)
    q.add_bond(3, 6, 4)
    q.add_bond(4, 5, 4)
    q.add_bond(5, 1, 4)
    q.add_bond(1, 6, 4)
    q.add_bond(6, 7, 4)
    q.add_bond(7, 8, 4)
    q.add_bond(8, 2, 4)
    rules.append((q, False))

    # N1C2=[NH+]C=CC2=CC=C1
    q = QueryContainer()
    q.add_atom('N', charge=1)  # first atom is charged
    q.add_atom('N')  # second is possible charged atom
    for _ in range(7):
        q.add_atom('A')
    q.add_bond(2, 7, 4)
    q.add_bond(7, 8, 4)
    q.add_bond(8, 9, 4)
    q.add_bond(9, 3, 4)
    q.add_bond(3, 4, 4)
    q.add_bond(3, 6, 4)
    q.add_bond(4, 5, 4)
    q.add_bond(5, 1, 4)
    q.add_bond(1, 6, 4)
    q.add_bond(6, 2, 4)
    rules.append((q, False))

    # N1C=C2C(=CC=[NH+]2)C=C1
    q = QueryContainer()
    q.add_atom('N', charge=1)  # first atom is charged
    q.add_atom('N')  # second is possible charged atom
    for _ in range(7):
        q.add_atom('A')
    q.add_bond(5, 2, 4)
    q.add_bond(5, 9, 4)
    q.add_bond(2, 3, 4)
    q.add_bond(3, 4, 4)
    q.add_bond(4, 1, 4)
    q.add_bond(4, 8, 4)
    q.add_bond(1, 6, 4)
    q.add_bond(6, 7, 4)
    q.add_bond(7, 8, 4)
    q.add_bond(8, 9, 4)
    rules.append((q, False))

    # C=1C=[NH+]C=2C=1NC=CC=2
    q = QueryContainer()
    q.add_atom('N', charge=1)  # first atom is charged
    q.add_atom('N')  # second is possible charged atom
    for _ in range(7):
        q.add_atom('A')
    q.add_bond(2, 6, 4)
    q.add_bond(2, 9, 4)
    q.add_bond(6, 3, 4)
    q.add_bond(3, 4, 4)
    q.add_bond(4, 5, 4)
    q.add_bond(5, 1, 4)
    q.add_bond(5, 9, 4)
    q.add_bond(1, 7, 4)
    q.add_bond(7, 8, 4)
    q.add_bond(8, 9, 4)
    rules.append((q, False))

    # C1=2C=[NH+]C=C1NC=CC=2
    q = QueryContainer()
    q.add_atom('N', charge=1)  # first atom is charged
    q.add_atom('N')  # second is possible charged atom
    for _ in range(7):
        q.add_atom('A')
    q.add_bond(6, 7, 4)
    q.add_bond(7, 8, 4)
    q.add_bond(8, 9, 4)
    q.add_bond(2, 9, 4)
    q.add_bond(2, 3, 4)
    q.add_bond(3, 4, 4)
    q.add_bond(3, 6, 4)
    q.add_bond(4, 1, 4)
    q.add_bond(5, 1, 4)
    q.add_bond(5, 6, 4)
    rules.append((q, False))

    # C1=2C=CNC=C1C=[NH+]C=2
    q = QueryContainer()
    q.add_atom('N', charge=1)  # first atom is charged
    q.add_atom('N')  # second is possible charged atom
    for _ in range(7):
        q.add_atom('A')
    q.add_bond(2, 6, 4)
    q.add_bond(2, 9, 4)
    q.add_bond(6, 3, 4)
    q.add_bond(3, 4, 4)
    q.add_bond(4, 5, 4)
    q.add_bond(4, 8, 4)
    q.add_bond(5, 1, 4)
    q.add_bond(1, 7, 4)
    q.add_bond(7, 8, 4)
    q.add_bond(8, 9, 4)
    rules.append((q, False))
    return rules


def _morgan_rules():
    """
    Rules working with Morgan.
    This rules is for charge canonization
    """
    from ...containers import QueryContainer

    rules: List[Tuple[QueryContainer, bool]] = []

    # N1C=CC2=CC=[NH+]N12
    q = QueryContainer()
    q.add_atom('N', charge=1)  # first atom is charged
    q.add_atom('N')  # second can be charged atom
    q.add_atom('N')
    for _ in range(5):
        q.add_atom('A')
    q.add_bond(1, 3, 4)
    q.add_bond(1, 4, 4)
    q.add_bond(3, 6, 4)
    q.add_bond(4, 5, 4)
    q.add_bond(5, 6, 4)
    q.add_bond(6, 7, 4)
    q.add_bond(7, 8, 4)
    q.add_bond(8, 2, 4)
    q.add_bond(2, 3, 4)
    rules.append((q, False))

    # N1C=CC2=[N+]1NC=C2
    q = QueryContainer()
    q.add_atom('N')  # first can be charged atom
    q.add_atom('N')  # second can be charged atom
    q.add_atom('N', charge=1)  # bad charged atom
    for _ in range(5):
        q.add_atom('A')
    q.add_bond(1, 3, 4)
    q.add_bond(1, 4, 4)
    q.add_bond(3, 6, 4)
    q.add_bond(4, 5, 4)
    q.add_bond(5, 6, 4)
    q.add_bond(6, 7, 4)
    q.add_bond(7, 8, 4)
    q.add_bond(8, 2, 4)
    q.add_bond(2, 3, 4)
    rules.append((q, True))

    # N1C=CN2C=C[NH+]=C12
    q = QueryContainer()
    q.add_atom('N', charge=1)  # first atom is charged
    q.add_atom('N')  # second can be charged atom
    q.add_atom('N')
    for _ in range(5):
        q.add_atom('A')
    q.add_bond(1, 4, 4)
    q.add_bond(2, 4, 4)
    q.add_bond(3, 4, 4)
    q.add_bond(2, 5, 4)
    q.add_bond(5, 6, 4)
    q.add_bond(6, 3, 4)
    q.add_bond(7, 8, 4)
    q.add_bond(8, 1, 4)
    q.add_bond(7, 3, 4)
    rules.append((q, False))

    # N1C=C[N+]2=C1NC=C2
    q = QueryContainer()
    q.add_atom('N')  # first can be charged atom
    q.add_atom('N')  # second can be charged atom
    q.add_atom('N', charge=1)  # bad charged atom
    for _ in range(5):
        q.add_atom('A')
    q.add_bond(1, 4, 4)
    q.add_bond(2, 4, 4)
    q.add_bond(3, 4, 4)
    q.add_bond(2, 5, 4)
    q.add_bond(5, 6, 4)
    q.add_bond(6, 3, 4)
    q.add_bond(7, 8, 4)
    q.add_bond(8, 1, 4)
    q.add_bond(7, 3, 4)
    rules.append((q, True))

    # imidazole
    q = QueryContainer()
    q.add_atom('N', charge=1)  # first atom is charged
    q.add_atom('N')  # second is possible charged atom
    for _ in range(3):
        q.add_atom('A')
    q.add_bond(1, 3, 4)
    q.add_bond(3, 2, 4)
    q.add_bond(2, 4, 4)
    q.add_bond(4, 5, 4)
    q.add_bond(5, 1, 4)
    rules.append((q, False))

    # pyrazole
    q = QueryContainer()
    q.add_atom('N', charge=1)  # first atom is charged
    q.add_atom('N')  # second is possible charged atom
    for _ in range(3):
        q.add_atom('A')
    q.add_bond(1, 2, 4)
    q.add_bond(2, 3, 4)
    q.add_bond(3, 4, 4)
    q.add_bond(4, 5, 4)
    q.add_bond(5, 1, 4)
    rules.append((q, False))
    return rules


fixed_rules = Proxy(_fixed_rules)
morgan_rules = Proxy(_morgan_rules)


__all__ = ['fixed_rules', 'morgan_rules']



# -*- coding: utf-8 -*-
#
#  Copyright 2014-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import namedtuple
from itertools import count
from ...containers import MoleculeContainer, ReactionContainer
from ...containers.bonds import Bond
from ...exceptions import AtomNotFound, MappingError
from ...periodictable import Element


parse_error = namedtuple('ParseError', ('number', 'position', 'log', 'meta'))


class Parser:
    """
    Override classes below then inheritance used.
    """
    MoleculeContainer = MoleculeContainer
    ReactionContainer = ReactionContainer

    def __init__(self, remap=False, ignore=False, store_log=False):
        self.__remap = remap
        self._ignore = ignore
        self._store_log = store_log
        self._log_buffer = []

    def _info(self, msg):
        self._log_buffer.append(msg)

    def _flush_log(self):
        self._log_buffer.clear()

    def _format_log(self):
        log = '\n'.join(self._log_buffer)
        self._log_buffer.clear()
        return log

    def _convert_reaction(self, data):
        if not (data['reactants'] or data['products'] or data['reagents']):
            raise ValueError('empty reaction')
        maps = {'reactants': [], 'products': [], 'reagents': []}
        for i, tmp in maps.items():
            for molecule in data[i]:
                used = set()
                for atom in molecule['atoms']:
                    m = atom['mapping']
                    if m:
                        if m in used:
                            if not self._ignore:
                                raise MappingError('mapping in molecules should be unique')
                            self._info(f'non-unique mapping in molecule: {m}')
                        else:
                            used.add(m)
                    tmp.append(m)

        length = count(max(max(maps['products'], default=0), max(maps['reactants'], default=0),
                           max(maps['reagents'], default=0)) + 1)

        # map unmapped atoms.
        for i, tmp in maps.items():
            used = set()
            maps[i] = remap = []
            for m in tmp:
                if not m:
                    remap.append(next(length))
                elif m in used:
                    if not self._ignore:
                        raise MappingError('mapping in reagents or products or reactants should be unique')
                    # force remap non unique atoms in molecules.
                    remap.append(next(length))
                    self._info(f'mapping in {i} changed from {m} to {remap[-1]}')
                else:
                    remap.append(m)
                    used.add(m)

        if maps['reagents']:
            tmp = (set(maps['reactants']) | set(maps['products'])) & set(maps['reagents'])
            if tmp:
                e = f'reagents has map intersection with reactants or products: {tmp}'
                if not self._ignore:
                    raise MappingError(e)
                self._info(e)
                maps['reagents'] = [x if x not in tmp else next(length) for x in maps['reagents']]

        # find breaks in map. e.g. 1,2,5,6. 3,4 - skipped
        if self.__remap:
            lose = sorted(set(range(1, next(length))) - set(maps['reactants']) - set(maps['products']) -
                          set(maps['reagents']), reverse=True)
            if lose:
                for i, tmp in maps.items():
                    if not tmp:
                        continue
                    for j in lose:
                        maps[i] = tmp = [x if x < j else x - 1 for x in tmp]

        rc = {'reactants': [], 'products': [], 'reagents': []}
        for i, tmp in maps.items():
            shift = 0
            for j in data[i]:
                atom_len = len(j['atoms'])
                remapped = {x: y for x, y in enumerate(tmp[shift: atom_len + shift])}
                shift += atom_len
                g = self._create_molecule(j, remapped)
                g.meta.update(j['meta'])
                rc[i].append(g)
        if self._store_log:
            if log := self._format_log():
                data['meta']['ParserLog'] = log
        else:
            self._flush_log()
        return ReactionContainer(meta=data['meta'], name=data.get('title'), **rc)

    def _convert_molecule(self, data):
        if self.__remap:
            remapped = {n: k for n, k in enumerate(range(1, len(data['atoms']) + 1))}
        else:
            length = count(max(x['mapping'] for x in data['atoms']) + 1)
            remapped, used = {}, set()
            for n, atom in enumerate(data['atoms']):
                m = atom['mapping']
                if not m:
                    remapped[n] = next(length)
                elif m in used:
                    if not self._ignore:
                        raise MappingError('mapping in molecules should be unique')
                    remapped[n] = next(length)
                    self._info(f'mapping in molecule changed from {m} to {remapped[n]}')
                else:
                    remapped[n] = m
                    used.add(m)
        mol = self._create_molecule(data, remapped)
        if self._store_log:
            if log := self._format_log():
                mol.meta['ParserLog'] = log
        else:
            self._flush_log()
        return mol

    def _create_molecule(self, data, mapping, *, _skip_calc_implicit=False):
        g = object.__new__(self.MoleculeContainer)
        pm = {}
        atoms = {}
        plane = {}
        charges = {}
        radicals = {}
        bonds = {}
        for n, atom in enumerate(data['atoms']):
            n = mapping[n]
            atoms[n] = Element.from_symbol(atom['element'])(atom['isotope'])
            bonds[n] = {}
            if (charge := atom['charge']) > 4 or charge < -4:
                raise ValueError('formal charge should be in range [-4, 4]')
            charges[n] = charge
            radicals[n] = atom['is_radical']
            plane[n] = (atom['x'], atom['y'])
            pm[n] = atom['mapping']
        for n, m, b in data['bonds']:
            n, m = mapping[n], mapping[m]
            if n == m:
                raise ValueError('atom loops impossible')
            if n not in bonds or m not in bonds:
                raise AtomNotFound('atoms not found')
            if n in bonds[m]:
                raise ValueError('atoms already bonded')
            bonds[n][m] = bonds[m][n] = Bond(b)
        if any(a['z'] for a in data['atoms']):
            conformers = [{mapping[n]: (a['x'], a['y'], a['z']) for n, a in enumerate(data['atoms'])}]
        else:
            conformers = []
        g.__setstate__(
                {'atoms': atoms, 'bonds': bonds, 'meta': data['meta'], 'plane': plane, 'parsed_mapping': pm,
                 'charges': charges, 'radicals': radicals, 'name': data.get('title', ''), 'conformers': conformers,
                 'atoms_stereo': {}, 'allenes_stereo': {}, 'cis_trans_stereo': {}, 'hydrogens': {}})
        if not _skip_calc_implicit:
            for n in atoms:
                g._calc_implicit(n)
        return g


__all__ = ['Parser', 'parse_error']



# -*- coding: utf-8 -*-
#
#  Copyright 2018-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from array import array
from CachedMethods import cached_args_method
from collections import defaultdict
from functools import cached_property
from itertools import chain, product
from typing import Dict, List, Set, Tuple, Union
from . import molecule  # cyclic imports resolve
from .bonds import Bond, QueryBond
from .graph import Graph
from ..algorithms.calculate2d import Calculate2DQuery
from ..algorithms.depict import DepictQuery
from ..algorithms.smiles import QuerySmiles
from ..algorithms.stereo import Stereo
from ..periodictable import AnyElement, AnyMetal, Element, ListElement, QueryElement
from ..periodictable.element import Query


class QueryContainer(Stereo, Graph[Query, QueryBond], QuerySmiles, DepictQuery, Calculate2DQuery):
    __slots__ = ('_neighbors', '_hybridizations', '_atoms_stereo', '_cis_trans_stereo', '_allenes_stereo',
                 '_hydrogens', '_rings_sizes', '_heteroatoms')

    _neighbors: Dict[int, Tuple[int, ...]]
    _hybridizations: Dict[int, Tuple[int, ...]]
    _atoms_stereo: Dict[int, bool]
    _allenes_stereo: Dict[int, bool]
    _cis_trans_stereo: Dict[Tuple[int, int], bool]
    _hydrogens: Dict[int, Tuple[int, ...]]
    _rings_sizes: Dict[int, Tuple[int, ...]]
    _heteroatoms: Dict[int, Tuple[int, ...]]

    def __init__(self):
        super().__init__()
        self._neighbors = {}
        self._hybridizations = {}
        self._atoms_stereo = {}
        self._allenes_stereo = {}
        self._cis_trans_stereo = {}
        self._hydrogens = {}
        self._rings_sizes = {}
        self._heteroatoms = {}

    def add_atom(self, atom: Union[Query, Element, int, str], *args,
                 neighbors: Union[int, List[int], Tuple[int, ...], None] = None,
                 hybridization: Union[int, List[int], Tuple[int, ...], None] = None,
                 hydrogens: Union[int, List[int], Tuple[int, ...], None] = None,
                 rings_sizes: Union[int, List[int], Tuple[int, ...], None] = None,
                 heteroatoms: Union[int, List[int], Tuple[int, ...], None] = None,
                 **kwargs):
        neighbors = self._validate_neighbors(neighbors)
        hybridization = self._validate_hybridization(hybridization)
        hydrogens = self._validate_neighbors(hydrogens)
        rings_sizes = self._validate_rings(rings_sizes)
        heteroatoms = self._validate_neighbors(heteroatoms)

        if not isinstance(atom, Query):
            if isinstance(atom, Element):
                atom = QueryElement.from_atomic_number(atom.atomic_number)(atom.isotope)
            elif isinstance(atom, str):
                atom = QueryElement.from_symbol(atom)()
            elif isinstance(atom, int):
                atom = QueryElement.from_atomic_number(atom)()
            else:
                raise TypeError('QueryElement object expected')

        _map = super().add_atom(atom, *args, **kwargs)
        self._neighbors[_map] = neighbors
        self._hybridizations[_map] = hybridization
        self._hydrogens[_map] = hydrogens
        self._rings_sizes[_map] = rings_sizes
        self._heteroatoms[_map] = heteroatoms
        return _map

    def add_bond(self, n, m, bond: Union[QueryBond, Bond, int, Tuple[int, ...]]):
        if isinstance(bond, Bond):
            bond = QueryBond.from_bond(bond)
        elif not isinstance(bond, QueryBond):
            bond = QueryBond(bond)

        sct = self._stereo_cis_trans_paths  # save
        sa = self._stereo_allenes_paths

        super().add_bond(n, m, bond)
        # remove stereo marks on bonded atoms and all its bonds
        if n in self._atoms_stereo:
            del self._atoms_stereo[n]
        if m in self._atoms_stereo:
            del self._atoms_stereo[m]
        if self._cis_trans_stereo:
            for nm, path in sct.items():
                if (n in path or m in path) and nm in self._cis_trans_stereo:
                    del self._cis_trans_stereo[nm]
        if self._allenes_stereo:
            for c, path in sa.items():
                if (n in path or m in path) and c in self._allenes_stereo:
                    del self._allenes_stereo[c]

    def delete_atom(self, n):
        bonds = set(self._bonds[n])  # save
        sct = self._stereo_cis_trans_paths
        sa = self._stereo_allenes_paths

        super().delete_atom(n)

        del self._neighbors[n]
        del self._hybridizations[n]
        del self._hydrogens[n]
        del self._rings_sizes[n]
        del self._heteroatoms[n]

        sas = self._atoms_stereo
        if n in sas:
            del sas[n]
        for m in bonds:
            if m in sas:
                del sas[m]
        if self._cis_trans_stereo:
            for nm, path in sct.items():
                if not bonds.isdisjoint(path) and nm in self._cis_trans_stereo:
                    del self._cis_trans_stereo[nm]
        if self._allenes_stereo:
            for c, path in sa.items():
                if not bonds.isdisjoint(path) and c in self._allenes_stereo:
                    del self._allenes_stereo[c]

    def delete_bond(self, n, m):
        sct = self._stereo_cis_trans_paths  # save
        sa = self._stereo_allenes_paths

        super().delete_bond(n, m)

        if n in self._atoms_stereo:
            del self._atoms_stereo[n]
        if m in self._atoms_stereo:
            del self._atoms_stereo[m]
        if self._cis_trans_stereo:
            for nm, path in sct.items():
                if (n in path or m in path) and nm in self._cis_trans_stereo:
                    del self._cis_trans_stereo[nm]
        if self._allenes_stereo:
            for c, path in sa.items():
                if (n in path or m in path) and c in self._allenes_stereo:
                    del self._allenes_stereo[c]

    def copy(self) -> 'QueryContainer':
        copy = super().copy()
        copy._neighbors = self._neighbors.copy()
        copy._hybridizations = self._hybridizations.copy()
        copy._hydrogens = self._hydrogens.copy()
        copy._heteroatoms = self._heteroatoms.copy()
        copy._rings_sizes = self._rings_sizes.copy()
        copy._atoms_stereo = self._atoms_stereo.copy()
        copy._allenes_stereo = self._allenes_stereo.copy()
        copy._cis_trans_stereo = self._cis_trans_stereo.copy()
        return copy

    def union(self, other: 'QueryContainer') -> 'QueryContainer':
        if not isinstance(other, QueryContainer):
            raise TypeError('QueryContainer expected')
        u = super().union(other)
        u._neighbors.update(other._neighbors)
        u._hybridizations.update(other._hybridizations)
        u._hydrogens.update(other._hydrogens)
        u._rings_sizes.update(other._rings_sizes)
        u._atoms_stereo.update(other._atoms_stereo)
        u._allenes_stereo.update(other._allenes_stereo)
        u._cis_trans_stereo.update(other._cis_trans_stereo)
        u._heteroatoms.update(other._heteroatoms)
        return u

    def get_mapping(self, other: Union['QueryContainer', 'molecule.MoleculeContainer'], /, *, _cython=True, **kwargs):
        # _cython - by default cython implementation enabled.
        # disable it by overriding method if Query Atoms or Containers logic changed.
        if isinstance(other, QueryContainer):
            return super().get_mapping(other, **kwargs)
        elif isinstance(other, molecule.MoleculeContainer):
            return super().get_mapping(other, _cython=_cython, **kwargs)
        raise TypeError('MoleculeContainer or QueryContainer expected')

    def enumerate_queries(self, *, enumerate_marks: bool = False):
        """
        Enumerate complex queries into multiple simple ones. For example `[N,O]-C` into `NC` and `OC`.

        :param enumerate_marks: enumerate multiple marks to separate queries
        """
        atoms = [(n, a._numbers) for n, a in self._atoms.items() if isinstance(a, ListElement)]
        bonds = [(n, m, b.order) for n, m, b in self.bonds() if len(b.order) > 1]
        for combo in product(*(x for *_, x in chain(atoms, bonds))):
            copy = self.copy()
            for (n, _), a in zip(atoms, combo):
                copy._atoms[n] = a = QueryElement.from_atomic_number(a)()
                a._attach_to_graph(copy, n)
            for (n, m, _), b in zip(bonds, combo[len(atoms):]):
                copy._bonds[n][m]._QueryBond__order = (b,)

            if enumerate_marks:
                c = 0
                slices = []
                data = []
                for attr in ('_neighbors', '_hybridizations', '_hydrogens', '_heteroatoms', '_rings_sizes'):
                    tmp = [(n, v) for n, v in getattr(self, attr).items() if len(v) > 1]
                    if tmp:
                        data.extend(tmp)
                        slices.append((attr, c, c + len(tmp)))
                        c += len(tmp)

                for combo2 in product(*(x for _, x in data)):
                    copy2 = copy.copy()
                    for attr, i, j in slices:
                        attr = getattr(copy2, attr)
                        for (n, _), v in zip(data[i: j], combo2[i: j]):
                            attr[n] = (v,)
                    yield copy2
            else:
                yield copy

    @cached_property
    def _screen_fingerprints(self) -> Tuple[Dict[int, Set[int]], ...]:
        """
        Fingerprints of all possible simple queries. Used for isomorphism tests filtering.
        """
        params = molecule.MoleculeContainer._fingerprint_config
        if not params or any(isinstance(a, (AnyElement, AnyMetal)) for _, a in self.atoms()):
            return ()  # skip queries with Any element

        fps = []
        for q in self.enumerate_queries():
            mol = molecule.MoleculeContainer()
            for n, a in q.atoms():
                mol.add_atom(Element.from_atomic_number(a.atomic_number)(a.isotope), n,
                             charge=a.charge, is_radical=a.is_radical)
            for n, m, b in q.bonds():
                mol.add_bond(n, m, Bond(b.order[0]))

            # extract only longest chains
            chains = []
            for c in sorted(mol._chains(**params), reverse=True, key=len):
                sc = set(c)
                if any(sc.issubset(x) for x in chains):
                    continue
                chains.append(c)
            if not chains:
                continue

            atoms = {idx: hash((atom.isotope or 0, atom.atomic_number, atom.charge, atom.is_radical))
                     for idx, atom in mol.atoms()}
            bonds = mol._bonds
            out = defaultdict(list)

            for frag in chains:
                var = [atoms[frag[0]]]
                for x, y in zip(frag, frag[1:]):
                    var.append(int(bonds[x][y]))
                    var.append(atoms[y])
                var = tuple(var)
                rev_var = var[::-1]
                out[hash(var if var > rev_var else rev_var)].extend(frag)

            fps.append({k: set(v) for k, v in out.items()})
        return tuple(fps)

    @cached_args_method
    def _component_fingerprints(self, component):
        """
        Fingerprints of specific component.
        """
        scope = set(self.connected_components[component])
        return tuple({k for k, v in fp.items() if not v.isdisjoint(scope)} for fp in self._screen_fingerprints)

    def _isomorphism_candidates(self, other, self_component, other_component):
        if self._screen_fingerprints and isinstance(other, molecule.MoleculeContainer):
            other_fingerprint = other._component_fingerprint(other_component)
            scope = set()
            for fp in self._component_fingerprints(self_component):
                if fp <= other_fingerprint.keys():
                    scope.update(x for k in fp for x in other_fingerprint[k])
            return scope
        return super()._isomorphism_candidates(other, self_component, other_component)

    @cached_property
    def _cython_compiled_query(self):
        # long I:
        # bond: single, double, triple, aromatic, special = 5 bit
        # atom: H-Ce: 58 bit
        # transfer bit

        # long II:
        # atom Pr-Og: 60 bit
        # hybridizations: 1-4 = 4 bit

        # long III:
        # isotope: not specified, isotope - common_isotope = -8 - +8 = 18 bit
        # is_radical: 2 bit
        # charge: -4 - +4: 9 bit
        # implicit_hydrogens: 0-4 = 5 bit
        # neighbors: 0-14 = 15 bit
        # heteroatoms: 0-14 = 15 bit

        # long IV:
        # ring_sizes: not-in-ring bit, 3-atom ring, 4-...., 65-atom ring
        from ..files._mdl.mol import common_isotopes

        _components, _closures = self._compiled_query
        components = []
        for c in _components:
            mapping = {n: i for i, (n, *_) in enumerate(c)}
            masks1 = []
            masks2 = []
            masks3 = []
            masks4 = []
            for *_, a, b in c:
                if isinstance(a, AnyMetal):  # isotope, radical, charge, hydrogens and heteroatoms states ignored
                    # elements except 1, 2, 5, 6, 7, 8, 9, 10, 14, 15, 16, 17, 18, 32, 33, 34, 35, 36, 51, 52, 53, 54
                    v1 = 0x0181c1fff07ffe1f
                    v2 = 0xfffffffffffffff0
                    v3 = 0xffffffffc0007fff
                    v4 = 0xffffffffffffffff
                else:
                    if isinstance(a, AnyElement):
                        v1 = 0x07ffffffffffffff
                        v2 = 0xfffffffffffffff0
                    else:
                        if isinstance(a, ListElement):
                            v1 = v2 = 0
                            for n in a._numbers:
                                if n > 58:
                                    v1 |= 1  # set transfer bit
                                    v2 |= 1 << (122 - n)
                                else:
                                    v1 |= 1 << (59 - n)
                        elif (n := a.atomic_number) > 58:
                            v1 = 1  # transfer bit
                            v2 = 1 << (122 - n)
                        else:
                            v1 = 1 << (59 - n)
                            v2 = 0
                    if a.isotope:
                        v3 = 1 << (a.isotope - common_isotopes[a.atomic_symbol] + 54)
                        if a.is_radical:
                            v3 |= 0x200000000000
                        else:
                            v3 |= 0x100000000000
                    elif a.is_radical:  # any isotope
                        v3 = 0xffffe00000000000
                    else:
                        v3 = 0xffffd00000000000

                    v3 |= 1 << (a.charge + 39)

                    if not a.implicit_hydrogens:
                        v3 |= 0x7c0000000
                    else:
                        for h in a.implicit_hydrogens:
                            v3 |= 1 << (h + 30)

                    if not a.heteroatoms:
                        v3 |= 0x7fff
                    else:
                        for n in a.heteroatoms:
                            v3 |= 1 << n

                    if a.ring_sizes:
                        if a.ring_sizes[0]:
                            v4 = 0
                            for r in a.ring_sizes:
                                if r > 65:  # big rings not supported
                                    continue
                                v4 |= 1 << (65 - r)
                            if not v4:  # only 65+ rings. set as rings-free.
                                v4 = 0x8000000000000000
                        else:  # not in rings
                            v4 = 0x8000000000000000
                    else:  # any rings
                        v4 = 0xffffffffffffffff

                if not a.neighbors:
                    v3 |= 0x3fff8000
                else:
                    for n in a.neighbors:
                        v3 |= 1 << (n + 15)

                if not a.hybridization:
                    v2 |= 0xf
                else:
                    for n in a.hybridization:
                        v2 |= 1 << (n - 1)

                if b is not None:
                    for o in b.order:
                        if o == 1:
                            v1 |= 0x0800000000000000
                        elif o == 4:
                            v1 |= 0x4000000000000000
                        elif o == 2:
                            v1 |= 0x1000000000000000
                        elif o == 3:
                            v1 |= 0x2000000000000000
                        else:
                            v1 |= 0x8000000000000000

                masks1.append(v1)
                masks2.append(v2)
                masks3.append(v3)
                masks4.append(v4)

            closures = [0] * len(c)  # closures amount
            q_from = [0] * len(c)
            q_to = [0] * len(c)
            indices = [0] * sum(len(ms) for n, ms in _closures.items() if n in mapping)
            bonds = indices.copy()

            start = 0
            for n, ms in _closures.items():
                if (i := mapping.get(n)) is not None:
                    closures[i] = len(ms)
                    q_from[i] = start
                    for j, (m, b) in enumerate(ms, start):
                        v = 0
                        for o in b.order:
                            if o == 1:
                                v |= 0x0800000000000000
                            elif o == 4:
                                v |= 0x4000000000000000
                            elif o == 2:
                                v |= 0x1000000000000000
                            elif o == 3:
                                v |= 0x2000000000000000
                            else:
                                v |= 0x8000000000000000
                        bonds[j] = v
                        indices[j] = mapping[m]
                    start += len(ms)
                    q_to[i] = start
            components.append((array('L', [n for n, *_ in c]), array('I', [0] + [mapping[x] for _, x, *_ in c[1:]]),
                               array('Q', masks1), array('Q', masks2), array('Q', masks3), array('Q', masks4),
                               array('I', closures), array('I', q_from), array('I', q_to),
                               array('I', indices), array('Q', bonds)))
        return components

    @staticmethod
    def _validate_neighbors(neighbors):
        if neighbors is None:
            neighbors = ()
        elif isinstance(neighbors, int):
            if neighbors < 0 or neighbors > 14:
                raise ValueError('neighbors should be in range [0, 14]')
            neighbors = (neighbors,)
        elif isinstance(neighbors, (tuple, list)):
            if not all(isinstance(n, int) for n in neighbors):
                raise TypeError('neighbors should be list or tuple of ints')
            if any(n < 0 or n > 14 for n in neighbors):
                raise ValueError('neighbors should be in range [0, 14]')
            if len(set(neighbors)) != len(neighbors):
                raise ValueError('neighbors should be unique')
            neighbors = tuple(sorted(neighbors))
        else:
            raise TypeError('neighbors should be int or list or tuple of ints')
        return neighbors

    @staticmethod
    def _validate_rings(rings):
        if rings is None:
            rings = ()
        elif isinstance(rings, int):
            if rings < 3 and rings != 0:
                raise ValueError('rings should be greater or equal 3. ring equal to zero is no ring atom mark')
            rings = (rings,)
        elif isinstance(rings, (tuple, list)):
            if not all(isinstance(n, int) for n in rings):
                raise TypeError('rings should be list or tuple of ints')
            if any(n < 3 for n in rings):
                raise ValueError('rings should be greater or equal 3')
            if len(set(rings)) != len(rings):
                raise ValueError('rings should be unique')
            rings = tuple(sorted(rings))
        else:
            raise TypeError('rings should be int or list or tuple of ints')
        return rings

    @staticmethod
    def _validate_hybridization(hybridization):
        if hybridization is None:
            hybridization = ()
        elif isinstance(hybridization, int):
            if hybridization < 1 or hybridization > 4:
                raise ValueError('hybridization should be in range [1, 4]')
            hybridization = (hybridization,)
        elif isinstance(hybridization, (tuple, list)):
            if not all(isinstance(h, int) for h in hybridization):
                raise TypeError('hybridizations should be list or tuple of ints')
            if any(h < 1 or h > 4 for h in hybridization):
                raise ValueError('hybridizations should be in range [1, 4]')
            if len(set(hybridization)) != len(hybridization):
                raise ValueError('hybridizations should be unique')
            hybridization = tuple(sorted(hybridization))
        else:
            raise TypeError('hybridization should be int or list or tuple of ints')
        return hybridization

    def __getstate__(self):
        return {'atoms_stereo': self._atoms_stereo, 'allenes_stereo': self._allenes_stereo,
                'cis_trans_stereo': self._cis_trans_stereo, 'neighbors': self._neighbors,
                'hybridizations': self._hybridizations, 'hydrogens': self._hydrogens,
                'rings_sizes': self._rings_sizes, 'heteroatoms': self._heteroatoms, **super().__getstate__()}

    def __setstate__(self, state):
        super().__setstate__(state)
        self._atoms_stereo = state['atoms_stereo']
        self._allenes_stereo = state['allenes_stereo']
        self._cis_trans_stereo = state['cis_trans_stereo']
        self._neighbors = state['neighbors']
        self._hybridizations = state['hybridizations']
        self._hydrogens = state['hydrogens']
        self._rings_sizes = state['rings_sizes']
        self._heteroatoms = state['heteroatoms']


__all__ = ['QueryContainer']



# -*- coding: utf-8 -*-
#
#  Copyright 2020, 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2020 Nail Samikaev <samikaevn@yandex.ru>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import deque, defaultdict
from functools import cached_property
from itertools import product, chain, repeat, combinations
from lazy_object_proxy import Proxy
from typing import TYPE_CHECKING, Iterator, Union, List
from ._acid import rules as acid_rules, stripped_rules as stripped_acid_rules
from ._base import rules as base_rules, stripped_rules as stripped_base_rules
from ...exceptions import InvalidAromaticRing
from ...periodictable import ListElement


if TYPE_CHECKING:
    from chython import MoleculeContainer


def _sugar_group():
    from ...containers import QueryContainer
    q = QueryContainer()
    q.add_atom(ListElement(['O', 'N']), hydrogens=(1, 2))  # enol
    q.add_atom(ListElement(['O', 'N']))  # ketone
    q.add_atom('C', hybridization=1)
    q.add_atom('C', hybridization=2)
    q.add_bond(1, 3, 1)
    q.add_bond(3, 4, 1)
    q.add_bond(2, 4, 2)
    return q


sugar_group = Proxy(_sugar_group)


class Tautomers:
    __slots__ = ()

    def neutralize(self: 'MoleculeContainer', *, fix_stereo=True, logging=False) -> Union[bool, List[int]]:
        """
        Convert organic salts to neutral form if possible. Only one possible form used for charge unbalanced structures.

        :param logging: return changed atoms list
        """
        try:
            mol, changed = next(self._neutralize())
        except StopIteration:
            if logging:
                return []
            return False

        self._charges.update(mol._charges)
        self._hydrogens.update(mol._hydrogens)
        self.flush_cache()
        if fix_stereo:
            self.fix_stereo()
        if logging:
            return changed
        return True

    def tautomerize(self: 'MoleculeContainer', *, prepare_molecules=True, limit: int = 1000) -> bool:
        """
        Convert structure to canonical tautomeric form. Return True if structure changed.
        """
        def key(m):
            # more aromatic rings is better
            # less charged atoms is better
            # lower Huckel energy is better
            # smiles alphabetically sorted (descent). bad rule!
            return len(m.aromatic_rings), -sum(x != 0 for x in m._charges.values()), -m.huckel_energy, str(m)

        # do zwitter-ions enumeration for charged structure - search canonic ion.
        canon = max(self.enumerate_tautomers(prepare_molecules=prepare_molecules, full=bool(int(self)), limit=limit),
                    key=key)
        if canon != self:  # attach state of canonic tautomer to self
            # atoms, radicals state, parsed_mapping and plane are unchanged
            self._bonds = canon._bonds
            self._charges = canon._charges  # for zwitter-ionic tautomers
            self._hydrogens = canon._hydrogens
            self._atoms_stereo = canon._atoms_stereo
            self._allenes_stereo = canon._allenes_stereo
            self._cis_trans_stereo = canon._cis_trans_stereo
            self._conformers.clear()  # flush 3d
            self.flush_cache()
            return True
        return False

    def enumerate_tautomers(self: 'MoleculeContainer', *, prepare_molecules=True, full=True, limit: int = 1000) -> \
            Iterator['MoleculeContainer']:
        """
        Enumerate all possible tautomeric forms of molecule.

        :param prepare_molecules: Standardize structures for correct processing.
        :param full: Do zwitter-ions enumeration.
        :param limit: Maximum amount of generated structures.
        """
        if limit < 3:
            raise ValueError('limit should be greater or equal 3')

        atoms_stereo = self._atoms_stereo
        allenes_stereo = self._allenes_stereo
        cis_trans_stereo = self._cis_trans_stereo

        # sssr, neighbors and heteroatoms are same for all tautomers.
        # prevent recalculation by sharing cache.
        if '__cached_args_method_neighbors' in self.__dict__:
            neighbors = self.__dict__['__cached_args_method_neighbors']
        else:
            neighbors = self.__dict__['__cached_args_method_neighbors'] = {}
        if '__cached_args_method_heteroatoms' in self.__dict__:
            heteroatoms = self.__dict__['__cached_args_method_heteroatoms']
        else:
            heteroatoms = self.__dict__['__cached_args_method_heteroatoms'] = {}

        has_stereo = bool(atoms_stereo or allenes_stereo or cis_trans_stereo)
        counter = 1

        copy = self.copy()
        copy.clean_stereo()
        # share cache
        copy.__dict__['sssr'] = self.sssr  # thiele
        copy.__dict__['ring_atoms'] = self.ring_atoms  # morgan
        copy.__dict__['connected_components'] = self.connected_components  # isomorphism
        copy.__dict__['atoms_rings_sizes'] = self.atoms_rings_sizes  # isomorphism
        copy.__dict__['__cached_args_method_neighbors'] = neighbors  # isomorphism
        copy.__dict__['__cached_args_method_heteroatoms'] = heteroatoms  # isomorphism
        if prepare_molecules:
            k = copy.kekule()
            i = copy.implicify_hydrogens(fix_stereo=False)
            if k or i:  # reset after flush
                copy.__dict__['sssr'] = self.sssr
                copy.__dict__['ring_atoms'] = self.ring_atoms
                copy.__dict__['connected_components'] = self.connected_components
                copy.__dict__['atoms_rings_sizes'] = self.atoms_rings_sizes
                copy.__dict__['__cached_args_method_neighbors'] = neighbors
                copy.__dict__['__cached_args_method_heteroatoms'] = heteroatoms

        thiele = copy.copy()
        thiele.__dict__['sssr'] = self.sssr
        thiele.__dict__['__cached_args_method_neighbors'] = neighbors
        if thiele.thiele(fix_tautomers=False):
            thiele.__dict__['sssr'] = self.sssr
            thiele.__dict__['__cached_args_method_neighbors'] = neighbors  # isomorphism
        thiele.__dict__['__cached_args_method_heteroatoms'] = heteroatoms  # isomorphism
        thiele.__dict__['ring_atoms'] = self.ring_atoms  # morgan
        thiele.__dict__['connected_components'] = self.connected_components  # isomorphism
        thiele.__dict__['atoms_rings_sizes'] = self.atoms_rings_sizes  # isomorphism

        if has_stereo:
            out = thiele.copy()
            out._atoms_stereo.update(atoms_stereo)
            out._allenes_stereo.update(allenes_stereo)
            out._cis_trans_stereo.update(cis_trans_stereo)
            out.fix_stereo()
            yield out  # no cache
        else:
            yield thiele

        seen = {thiele: None}  # value is parent molecule - required for preventing migrations in sugars.

        # first of all try to neutralize
        if copy.neutralize(fix_stereo=False):
            thiele = copy.copy()

            copy.__dict__['sssr'] = self.sssr
            copy.__dict__['ring_atoms'] = self.ring_atoms
            copy.__dict__['connected_components'] = self.connected_components
            copy.__dict__['atoms_rings_sizes'] = self.atoms_rings_sizes
            copy.__dict__['__cached_args_method_neighbors'] = neighbors
            copy.__dict__['__cached_args_method_heteroatoms'] = heteroatoms

            thiele.__dict__['sssr'] = self.sssr
            thiele.__dict__['__cached_args_method_neighbors'] = neighbors
            if thiele.thiele(fix_tautomers=False):
                thiele.__dict__['sssr'] = self.sssr
                thiele.__dict__['__cached_args_method_neighbors'] = neighbors
            thiele.__dict__['__cached_args_method_heteroatoms'] = heteroatoms
            thiele.__dict__['ring_atoms'] = self.ring_atoms
            thiele.__dict__['connected_components'] = self.connected_components
            thiele.__dict__['atoms_rings_sizes'] = self.atoms_rings_sizes

            if has_stereo:
                out = thiele.copy()
                out._atoms_stereo.update(atoms_stereo)
                out._allenes_stereo.update(allenes_stereo)
                out._cis_trans_stereo.update(cis_trans_stereo)
                out.fix_stereo()
                yield out
            else:
                yield thiele
            counter += 1
            seen[thiele] = None

        # lets iteratively do keto-enol transformations.
        rings_count = len(thiele.aromatic_rings)  # increase rings strategy.
        queue = deque([(copy, thiele)])
        new_queue = [thiele]  # new_queue - molecules suitable for hetero-arenes enumeration.
        # store aromatic form to seen. kekule forms not suitable for duplicate checking.

        while queue:
            current, thiele_current = queue.popleft()
            for mol, ket in current._enumerate_keto_enol_tautomers():
                thiele = mol.copy()
                # set cache to child structures.
                mol.__dict__['sssr'] = self.sssr
                mol.__dict__['ring_atoms'] = self.ring_atoms
                mol.__dict__['connected_components'] = self.connected_components
                mol.__dict__['atoms_rings_sizes'] = self.atoms_rings_sizes
                mol.__dict__['__cached_args_method_neighbors'] = neighbors
                mol.__dict__['__cached_args_method_heteroatoms'] = heteroatoms

                thiele.__dict__['sssr'] = self.sssr
                thiele.__dict__['__cached_args_method_neighbors'] = neighbors
                if thiele.thiele(fix_tautomers=False):  # reset cache after flush_cache.
                    thiele.__dict__['sssr'] = self.sssr
                    thiele.__dict__['__cached_args_method_neighbors'] = neighbors
                thiele.__dict__['__cached_args_method_heteroatoms'] = heteroatoms
                thiele.__dict__['ring_atoms'] = self.ring_atoms
                thiele.__dict__['connected_components'] = self.connected_components
                thiele.__dict__['atoms_rings_sizes'] = self.atoms_rings_sizes

                if thiele not in seen:
                    seen[thiele] = current
                    rc = len(thiele.aromatic_rings)
                    if rc < rings_count:  # skip aromatic rings destruction
                        continue
                    elif rc > rings_count:  # higher aromaticity found. flush old queues.
                        rings_count = rc
                        queue = deque([(mol, thiele)])
                        new_queue = [thiele]
                        copy = mol  # new entry point.
                        if has_stereo:
                            thiele = thiele.copy()
                            thiele._atoms_stereo.update(atoms_stereo)
                            thiele._allenes_stereo.update(allenes_stereo)
                            thiele._cis_trans_stereo.update(cis_trans_stereo)
                            thiele.fix_stereo()
                        yield thiele
                        break
                    elif current is not copy and ket:  # prevent carbonyl migration in sugars. skip entry point.
                        # search alpha hydroxy ketone inversion
                        before = seen[thiele_current]._sugar_groups
                        if any((k, e) in before for e, k in mol._sugar_groups):
                            continue

                    queue.append((mol, thiele))
                    new_queue.append(thiele)
                    if has_stereo:
                        thiele = thiele.copy()
                        thiele._atoms_stereo.update(atoms_stereo)
                        thiele._allenes_stereo.update(allenes_stereo)
                        thiele._cis_trans_stereo.update(cis_trans_stereo)
                        thiele.fix_stereo()
                    yield thiele
                    counter += 1
                    if counter == limit:
                        return

        queue = deque(new_queue)
        while queue:
            current = queue.popleft()
            for mol in current._enumerate_hetero_arene_tautomers():
                mol.__dict__['sssr'] = self.sssr
                mol.__dict__['__cached_args_method_neighbors'] = neighbors
                mol.__dict__['__cached_args_method_heteroatoms'] = heteroatoms
                mol.__dict__['ring_atoms'] = self.ring_atoms
                mol.__dict__['connected_components'] = self.connected_components
                mol.__dict__['atoms_rings_sizes'] = self.atoms_rings_sizes
                if mol not in seen:
                    seen[mol] = None
                    queue.append(mol)
                    new_queue.append(mol)  # new hetero-arenes also should be included to this list.
                    if has_stereo:
                        mol = mol.copy()
                        mol._atoms_stereo.update(atoms_stereo)
                        mol._allenes_stereo.update(allenes_stereo)
                        mol._cis_trans_stereo.update(cis_trans_stereo)
                        mol.fix_stereo()
                    yield mol
                    counter += 1
                    if counter == limit:
                        return

        if not full:
            return

        queue = deque(new_queue)
        while queue:
            current = queue.popleft()
            for mol in current._enumerate_zwitter_tautomers():
                mol.__dict__['sssr'] = self.sssr
                mol.__dict__['__cached_args_method_neighbors'] = neighbors
                mol.__dict__['__cached_args_method_heteroatoms'] = heteroatoms
                mol.__dict__['ring_atoms'] = self.ring_atoms
                mol.__dict__['connected_components'] = self.connected_components
                mol.__dict__['atoms_rings_sizes'] = self.atoms_rings_sizes
                if mol not in seen:
                    seen[mol] = None
                    queue.append(mol)
                    if has_stereo:
                        mol = mol.copy()  # prevent seen hashtable breaking
                        mol._atoms_stereo.update(atoms_stereo)
                        mol._allenes_stereo.update(allenes_stereo)
                        mol._cis_trans_stereo.update(cis_trans_stereo)
                        mol.fix_stereo()
                    yield mol
                    counter += 1
                    if counter == limit:
                        return

    def _neutralize(self: 'MoleculeContainer'):
        donors = []
        acceptors = []
        for q in stripped_acid_rules:
            for mapping in q.get_mapping(self, automorphism_filter=False):
                donors.append(mapping[1])
        for q in stripped_base_rules:
            for mapping in q.get_mapping(self, automorphism_filter=False):
                acceptors.append(mapping[1])

        if not donors or not acceptors:
            return
        elif len(donors) > len(acceptors):
            copy = self.copy()
            for a in acceptors:
                copy._hydrogens[a] += 1
                copy._charges[a] += 1
            for c in combinations(donors, len(acceptors)):
                mol = copy.copy()
                for d in c:
                    mol._hydrogens[d] -= 1
                    mol._charges[d] -= 1
                yield mol, acceptors + list(c)
        elif len(donors) < len(acceptors):
            copy = self.copy()
            for d in donors:
                copy._hydrogens[d] -= 1
                copy._charges[d] -= 1
            for c in combinations(acceptors, len(donors)):
                mol = copy.copy()
                for a in c:
                    mol._hydrogens[a] += 1
                    mol._charges[a] += 1
                yield mol, donors + list(c)
        else:
            mol = self.copy()
            for d in donors:
                mol._hydrogens[d] -= 1
                mol._charges[d] -= 1
            for a in acceptors:
                mol._hydrogens[a] += 1
                mol._charges[a] += 1
            yield mol, donors + acceptors

    def _enumerate_keto_enol_tautomers(self: Union['MoleculeContainer', 'Tautomers']):
        for fix, ket in self.__enumerate_bonds():
            if ket:
                a = fix[-1][1]
                d = fix[0][0]
            else:
                a = fix[0][0]
                d = fix[-1][1]

            mol = self.copy()
            m_bonds = mol._bonds
            for n, m, b in fix:
                m_bonds[n][m]._Bond__order = b

            mol._hydrogens[a] += 1
            mol._hydrogens[d] -= 1
            yield mol, ket

    def _enumerate_zwitter_tautomers(self: 'MoleculeContainer'):
        donors = []
        acceptors = []
        for q in acid_rules:
            for mapping in q.get_mapping(self, automorphism_filter=False):
                donors.append(mapping[1])
        for q in base_rules:
            for mapping in q.get_mapping(self, automorphism_filter=False):
                acceptors.append(mapping[1])

        for d, a in product(donors, acceptors):
            mol = self.copy()
            mol._hydrogens[d] -= 1
            mol._hydrogens[a] += 1
            mol._charges[d] -= 1
            mol._charges[a] += 1
            yield mol

    def _enumerate_hetero_arene_tautomers(self: 'MoleculeContainer'):
        atoms = self._atoms
        bonds = self._bonds
        hydrogens = self._hydrogens
        charges = self._charges
        radicals = self._radicals

        rings = defaultdict(list)  # aromatic skeleton
        for n, m_bond in bonds.items():
            for m, bond in m_bond.items():
                if bond.order == 4:
                    rings[n].append(m)
        if not rings:
            return

        acceptors = set()
        donors = set()
        single_bonded = set()
        for n, ms in rings.items():
            if len(ms) == 2:
                if atoms[n].atomic_number in (5, 7, 15) and not charges[n] and not radicals[n]:
                    # only neutral B, N, P
                    if hydrogens[n]:  # pyrrole
                        donors.add(n)
                    elif len(bonds[n]) == 2:  # pyridine
                        acceptors.add(n)
                    else:
                        single_bonded.add(n)
                elif charges[n] == -1 and atoms[n].atomic_number == 6:  # ferrocene
                    single_bonded.add(n)
            elif len(ms) == 3 and atoms[n].atomic_number in (5, 7, 15) and not charges[n] and not radicals[n]:
                single_bonded.add(n)
        if not donors or not acceptors:
            return

        atoms = set(rings)
        components = []
        while atoms:
            start = atoms.pop()
            component = {start: rings[start]}
            queue = deque([start])
            while queue:
                current = queue.popleft()
                for n in rings[current]:
                    if n not in component:
                        queue.append(n)
                        component[n] = rings[n]

            atoms.difference_update(component)
            if donors.isdisjoint(component) or acceptors.isdisjoint(component):
                continue
            components.append(component)

        if not components:
            return
        for component in components:
            for d, a in product(component.keys() & donors, component.keys() & acceptors):
                sb = component.keys() & single_bonded
                sb.add(a)  # now pyrrole
                try:
                    next(self._kekule_component(component, sb, ()))
                except InvalidAromaticRing:
                    continue
                mol = self.copy()
                mol._hydrogens[d] = 0
                mol._hydrogens[a] = 1
                yield mol

    @cached_property
    def _sugar_groups(self):
        ek = []
        for mapping in sugar_group.get_mapping(self, automorphism_filter=False):
            e, k = mapping[1], mapping[2]
            ek.append((e, k))
        return ek

    def __enumerate_bonds(self: 'MoleculeContainer'):
        atoms = self._atoms
        bonds = self._bonds
        charge = self._charges
        hydrogens = self._hydrogens
        hybridization = self.hybridization
        heteroatoms = self.heteroatoms
        neighbors = self.neighbors

        # search neutral oxygen and nitrogen
        donors = []
        acceptors = []
        for n, a in atoms.items():
            if charge[n] or not neighbors(n) or heteroatoms(n):
                continue
            if a.atomic_number in (7, 8):
                if hybridization(n) == 2:  # imine!
                    acceptors.append(n)
                elif hydrogens[n]:  # amine!
                    donors.append(n)

        for atom, hydrogen in chain(zip(donors, repeat(True)), zip(acceptors, repeat(False))):
            path = []
            seen = {atom}
            # stack is neighbors
            if hydrogen:  # enol
                stack = [(atom, n, 2, 0) for n in bonds[atom] if hybridization(n) == 2]
            else:  # ketone
                stack = [(atom, n, 1, 0) for n, b in bonds[atom].items() if hybridization(n) == 2 and b.order == 2]

            while stack:
                last, current, bond, depth = stack.pop()

                if len(path) > depth:
                    if len(path) % 2 == 0:
                        yield path, hydrogen
                    seen.difference_update(x for _, x, _ in path[depth:])
                    path = path[:depth]

                path.append((last, current, bond))

                # adding neighbors
                depth += 1
                seen.add(current)
                if bond == 2:
                    next_bond = 1
                else:
                    next_bond = 2

                for n, b in bonds[current].items():
                    if n == last:
                        continue
                    elif n in seen:  # aromatic ring destruction. pyridine double bonds shift
                        if not stack:
                            path = []
                    elif b.order == bond and atoms[n].atomic_number == 6:
                        hb = hybridization(n)
                        if hb == 2:  # grow up
                            stack.append((current, n, next_bond, depth))
                        elif hydrogen:
                            if hb == 3:  # OC=CC=C=A case
                                cp = path.copy()
                                cp.append((current, n, 1))
                                yield cp, True  # ketone found
                        elif hb == 1 and hydrogens[n]:  # ketone >> enole
                            cp = path.copy()
                            cp.append((current, n, 2))
                            yield cp, False

            if path and len(path) % 2 == 0:  # time to yield
                yield path, hydrogen


__all__ = ['Tautomers']



# -*- coding: utf-8 -*-
#
#  Copyright 2017-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2019 Timur Gimadiev <timur.gimadiev@gmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from abc import ABC, abstractmethod
from CachedMethods import cached_method
from collections import defaultdict
from functools import cached_property
from hashlib import sha512
from itertools import count, product
from random import random
from typing import Callable, Tuple, TYPE_CHECKING, Union


if TYPE_CHECKING:
    from chython import MoleculeContainer, CGRContainer, QueryContainer
    from chython.containers.graph import Graph

charge_str = {-4: '-4', -3: '-3', -2: '-2', -1: '-', 0: '0', 1: '+', 2: '+2', 3: '+3', 4: '+4'}
order_str = {1: '-', 2: '=', 3: '#', 4: ':', 8: '~', None: '.'}
organic_set = {'C', 'N', 'O', 'P', 'S', 'F', 'Cl', 'Br', 'I', 'B'}
hybridization_str = {4: 'a', 3: 't', 2: 'd', 1: 's', None: 'n'}
dyn_order_str = {(None, 1): '[.>-]', (None, 2): '[.>=]', (None, 3): '[.>#]', (None, 4): '[.>:]', (None, 8): '[.>~]',
                 (1, None): '[->.]', (1, 1): '', (1, 2): '[->=]', (1, 3): '[->#]', (1, 4): '[->:]', (1, 8): '[->~]',
                 (2, None): '[=>.]', (2, 1): '[=>-]', (2, 2): '=', (2, 3): '[=>#]', (2, 4): '[=>:]', (2, 8): '[=>~]',
                 (3, None): '[#>.]', (3, 1): '[#>-]', (3, 2): '[#>=]', (3, 3): '#', (3, 4): '[#>:]', (3, 8): '[#>~]',
                 (4, None): '[:>.]', (4, 1): '[:>-]', (4, 2): '[:>=]', (4, 3): '[:>#]', (4, 4): ':', (4, 8): '[:>~]',
                 (8, None): '[~>.]', (8, 1): '[~>-]', (8, 2): '[~>=]', (8, 3): '[~>#]', (8, 4): '[~>:]', (8, 8): '~'}

dyn_charge_str = {(i, j): f'{charge_str[i]}>{charge_str[j]}' if i != j else charge_str[i]
                  for i, j in product(range(-4, 5), repeat=2)}
dyn_charge_str[(0, 0)] = ''

dyn_radical_str = {(True, True): '*', (True, False): '*>^', (False, True): '^>*'}


class Smiles(ABC):
    __slots__ = ()

    @cached_method
    def __str__(self):
        return ''.join(self._smiles(self._smiles_order()))

    def __format__(self: Union['Graph', 'Smiles'], format_spec):
        """
        Signature generation options.

        :param format_spec: String with keys:
            a - Generate asymmetric closures.
            !s - Disable stereo marks.
            A - Use aromatic bonds instead aromatic atoms.
            m - Set atom mapping.
            r - Generate random-ordered smiles.
            h - Show implicit hydrogens.
            !b - Disable bonds tokens.

            Combining possible. Order independent. Another keys ignored.
        """
        if format_spec:
            kwargs = {}
            if 'a' in format_spec:
                kwargs['asymmetric_closures'] = True
            if '!s' in format_spec:
                kwargs['stereo'] = False
            if 'A' in format_spec:
                kwargs['aromatic'] = False
            if 'm' in format_spec:
                kwargs['mapping'] = True
            if 'h' in format_spec:
                kwargs['hydrogens'] = True
            if '!b' in format_spec:
                kwargs['bonds'] = False
            if 'r' in format_spec:
                kwargs['random'] = True

                def w(_):
                    return random()
            else:
                w = self._smiles_order('!s' not in format_spec)
            return ''.join(self._smiles(w, **kwargs))
        return str(self)

    def __eq__(self, other):
        return isinstance(other, Smiles) and str(self) == str(other)

    @cached_method
    def __hash__(self):
        return hash(str(self))

    @cached_method
    def __bytes__(self):
        return sha512(str(self).encode()).digest()

    @cached_property
    def smiles_atoms_order(self) -> Tuple[int, ...]:
        """
        Atoms order in canonic SMILES.
        """
        smiles, order = self._smiles(self._smiles_order(), _return_order=True)
        self.__dict__['__cached_method___str__'] = ''.join(smiles)  # cache smiles also
        return tuple(order)

    def _smiles(self: Union['Graph', 'Smiles'], weights, *, asymmetric_closures=False,
                open_parenthesis='(', close_parenthesis=')', delimiter='.', _return_order=False, **kwargs):
        if not self._atoms:
            return []
        bonds = self._bonds
        atoms_set = set(self._atoms)
        seen = {}
        cycles = count()
        casted_cycles = {}
        string = []
        order = []
        visited_bond = set()

        groups = defaultdict(int)
        for n in atoms_set:
            groups[weights(n)] -= 1

        if kwargs.get('random', False):
            mod_weights_start = mod_weights = weights
        else:
            def mod_weights_start(x):
                return (groups[weights(x)],  # common groups
                        weights(x))  # smallest weight

            def mod_weights(x):
                return (groups[weights(x)],  # common groups
                        weights(x),  # smallest weight
                        seen[x])  # BFS nearest to starting

        while True:
            start = min(atoms_set, key=mod_weights_start)
            seen[start] = 0
            queue = [(start, 1)]
            while queue:
                n, d = queue.pop(0)
                for m in bonds[n].keys() - seen.keys():
                    queue.append((m, d + 1))
                    seen[m] = d

            # modified NX dfs with cycle detection
            stack = [(start, len(atoms_set), iter(sorted(bonds[start], key=mod_weights)))]
            visited = {start: []}  # predecessors for stereo. atom: (visited[atom], *edges[atom])
            disconnected = set()
            edges = defaultdict(list)
            tokens = defaultdict(list)
            while stack:
                parent, depth_now, children = stack[-1]
                try:
                    child = next(children)
                except StopIteration:
                    stack.pop()
                else:
                    if child not in visited:
                        edges[parent].append(child)
                        visited[child] = [parent]
                        if depth_now > 1:
                            front = bonds[child].keys() - {parent}
                            if front:
                                stack.append((child, depth_now - 1, iter(sorted(front, key=mod_weights))))
                    elif (child, parent) not in disconnected:
                        disconnected.add((parent, child))
                        disconnected.add((child, parent))
                        cycle = next(cycles)
                        tokens[parent].append((child, cycle))
                        tokens[child].append((parent, cycle))

            # flatten directed graph: edges
            stack = [[start, 0, [start]]]
            while True:
                tail, closure, smiles = stack[-1]
                if tail in edges:
                    children = edges[tail]
                    if len(children) > 1:  # has side chain
                        child = children[-1]
                        stack_len = len(stack)
                        stack.append([child, 0, [(tail, child), child]])  # end of current chain
                        for child in children[-2::-1]:  # start side chains
                            stack.append([child, stack_len, ['(', (tail, child), child]])
                    else:  # chain grow
                        child = children[-1]
                        stack[-1][0] = child
                        smiles.append((tail, child))
                        smiles.append(child)
                elif closure:  # end of side chain
                    stack.pop()
                    if smiles[-2] == '(':
                        smiles.pop(-2)
                    else:
                        smiles.append(')')
                    stack[closure - 1][2].extend(smiles)
                elif len(stack) > 2:
                    stack.pop()
                    stack[-1][0] = tail
                    stack[-1][2].extend(smiles)
                elif len(stack) == 2:
                    stack[0][2].extend(smiles)
                    smiles = stack[0][2]
                    break
                else:
                    break

            # prepare new neighbors order for stereo sign calculation
            for token in smiles:
                if token in tokens:
                    tokens[token].sort(key=lambda x: casted_cycles.get(x[1]) or
                                                     casted_cycles.setdefault(x[1], len(casted_cycles) + 1))
                    visited[token].extend(n for n, _ in tokens[token])
                if token in edges:
                    visited[token].extend(edges[token])

            for token in smiles:
                if isinstance(token, int):  # atoms
                    string.append(self._format_atom(token, visited, **kwargs))
                    order.append(token)
                    if token in tokens:
                        for m, c in tokens[token]:
                            if asymmetric_closures:
                                if (token, m) not in visited_bond:
                                    string.append(self._format_bond(token, m, visited, **kwargs))
                                    visited_bond.add((m, token))
                            else:
                                string.append(self._format_bond(token, m, visited, **kwargs))
                            string.append(self._format_closure(casted_cycles[c]))
                elif token == '(':
                    string.append(open_parenthesis)
                elif token == ')':
                    string.append(close_parenthesis)
                else:  # bonds
                    string.append(self._format_bond(*token, visited, **kwargs))

            atoms_set.difference_update(visited)
            if atoms_set:
                string.append(delimiter)
            else:
                break
        if _return_order:
            return string, order
        return string

    @staticmethod
    def _format_closure(c):
        return str(c) if c < 10 else f'%{c}'

    @abstractmethod
    def _format_atom(self, n, adjacency, **kwargs):
        ...

    @abstractmethod
    def _format_bond(self, n, m, adjacency, **kwargs):
        ...

    def _smiles_order(self: 'Graph', stereo=True) -> Callable:
        return self.atoms_order.__getitem__


class MoleculeSmiles(Smiles):
    __slots__ = ()

    def _smiles_order(self: 'MoleculeContainer', stereo=True) -> Callable:
        if stereo:
            return self._chiral_morgan.__getitem__
        else:
            return self.atoms_order.__getitem__

    def _format_atom(self: 'MoleculeContainer', n, adjacency, **kwargs):
        atom = self._atoms[n]
        charge = self._charges[n]
        ih = self._hydrogens[n]
        hyb = self.hybridization(n)

        smi = ['',  # [
               str(atom.isotope) if atom.isotope else '',  # isotope
               None,
               '',  # stereo
               '',  # hydrogen
               '',  # charge
               f':{n}' if kwargs.get('mapping', False) else '',  # mapping
               '']  # ]

        if kwargs.get('stereo', True):
            if n in self._atoms_stereo:
                if ih and next(x for x in adjacency) == n:  # first atom in smiles has reversed chiral mark
                    smi[3] = '@@' if self._translate_tetrahedron_sign(n, adjacency[n]) else '@'
                else:
                    smi[3] = '@' if self._translate_tetrahedron_sign(n, adjacency[n]) else '@@'
            elif n in self._allenes_stereo:
                t1, t2 = self._stereo_allenes_terminals[n]
                env = self._stereo_allenes[n]
                n1 = next(x for x in adjacency[t1] if x in env)
                n2 = next(x for x in adjacency[t2] if x in env)
                smi[3] = '@' if self._translate_allene_sign(n, n1, n2) else '@@'
            elif charge:
                smi[5] = charge_str[charge]
        elif charge:
            smi[5] = charge_str[charge]

        if any(smi) or atom.atomic_symbol not in organic_set or self._radicals[n] or kwargs.get('hydrogens', False):
            smi[0] = '['
            smi[-1] = ']'
            if ih == 1:
                smi[4] = 'H'
            elif ih:
                smi[4] = f'H{ih}'
        elif hyb == 4 and ih and atom.atomic_number in (5, 7, 15):  # pyrole
            smi[0] = '['
            smi[-1] = ']'
            if ih == 1:
                smi[4] = 'H'
            else:
                smi[4] = f'H{ih}'

        if kwargs.get('aromatic', True) and hyb == 4:
            smi[2] = atom.atomic_symbol.lower()
        else:
            smi[2] = atom.atomic_symbol
        return ''.join(smi)

    def _format_bond(self: 'MoleculeContainer', n, m, adjacency, **kwargs):
        if not kwargs.get('bonds', True):
            return ''
        order = self._bonds[n][m].order
        if order == 4:
            if kwargs.get('aromatic', True):
                return ''
            return ':'
        elif kwargs.get('stereo', True) and order == 1:  # cis-trans /\
            ctt = self._stereo_cis_trans_terminals
            if n in ctt:
                ts = ctt[n]
                if ts in self._cis_trans_stereo:
                    env = self._stereo_cis_trans[ts]
                    if m == next(x for x in adjacency[n] if x in env):  # only first neighbor of double bonded atom
                        if n == next(x for x in adjacency if x in ts):  # Cn(\Rm)(X)=C, Cn(=C)(\Rm)X, C(=C=Cn(\Rm)X)=C
                            return '\\'
                        else:  # C=Cn(Rm)(X) cases
                            n2 = ts[1] if ts[0] == n else ts[0]
                            m2 = next(x for x in adjacency[n2] if x in env)
                            return '\\' if self._translate_cis_trans_sign(n2, n, m2, m) else '/'
            elif m in ctt and self._atoms[n].atomic_number != 1:  # Hn-Cm= case ignored!
                ts = ctt[m]
                if ts in self._cis_trans_stereo:
                    if m == next(x for x in adjacency if x in ts):  # Rn-Cm(X)=C case
                        return '/'  # always start with UP R/C=C-X. RUSSIANS POSITIVE!
                    else:  # second RnCm=1X or R1...=C1(X) case
                        env = self._stereo_cis_trans[ts]
                        n2 = ts[1] if ts[0] == m else ts[0]
                        m2 = next(x for x in adjacency[n2] if x in env)
                        return '/' if self._translate_cis_trans_sign(n2, m, m2, n) else '\\'

            if kwargs.get('aromatic', True) and self.hybridization(n) == self.hybridization(m) == 4:
                return '-'
            return ''
        elif order == 2:
            return '='
        elif order == 3:
            return '#'
        elif order == 1:
            if kwargs.get('aromatic', True) and self.hybridization(n) == self.hybridization(m) == 4:
                return '-'
            return ''
        else:  # order == 8
            return '~'


class CGRSmiles(Smiles):
    __slots__ = ()

    def _format_atom(self: 'CGRContainer', n, adjacency, **kwargs):
        atom = self._atoms[n]
        charge = self._charges[n]
        is_radical = self._radicals[n]
        p_charge = self._p_charges[n]
        p_is_radical = self._p_radicals[n]
        if atom.isotope:
            smi = [str(atom.isotope), atom.atomic_symbol]
        else:
            smi = [atom.atomic_symbol]

        if charge or p_charge:
            smi.append(dyn_charge_str[(charge, p_charge)])
        if is_radical or p_is_radical:
            smi.append(dyn_radical_str[(is_radical, p_is_radical)])

        if len(smi) != 1 or atom.atomic_symbol not in organic_set:
            smi.insert(0, '[')
            smi.append(']')
        return ''.join(smi)

    def _format_bond(self: 'CGRContainer', n, m, adjacency, **kwargs):
        bond = self._bonds[n][m]
        return dyn_order_str[(bond.order, bond.p_order)]


class QuerySmiles(Smiles):
    __slots__ = ()

    def _format_atom(self: 'QueryContainer', n, adjacency, **kwargs):
        atom = self._atoms[n]
        charge = self._charges[n]
        hybridization = self._hybridizations[n]
        neighbors = self._neighbors[n]
        hydrogens = self._hydrogens[n]
        rings = self._rings_sizes[n]
        heteroatoms = self._heteroatoms[n]

        if atom.isotope:
            smi = ['[', str(atom.isotope), atom.atomic_symbol]
        else:
            smi = ['[', atom.atomic_symbol]

        if n in self._atoms_stereo:  # carbon only
            smi.append('@' if self._translate_tetrahedron_sign(n, adjacency[n]) else '@@')

        if neighbors:
            smi.append(';D')
            smi.append(''.join(str(x) for x in neighbors))

        if hydrogens:
            smi.append(';H')
            smi.append(''.join(str(x) for x in hydrogens))

        if rings:
            smi.append(';r')
            smi.append(''.join(str(x) for x in rings))

        if hybridization:
            smi.append(';Z')
            smi.append(''.join(hybridization_str[x] for x in hybridization))

        if heteroatoms:
            smi.append(';W')
            smi.append(''.join(str(x) for x in heteroatoms))

        if self._radicals[n]:
            smi.append(';*')

        if charge:
            smi.append(charge_str[charge])

        smi.append(']')
        return ''.join(smi)

    def _format_bond(self: 'QueryContainer', n, m, adjacency, **kwargs):
        return ','.join(order_str[x] for x in self._bonds[n][m].order)


__all__ = ['MoleculeSmiles', 'CGRSmiles', 'QuerySmiles']



# -*- coding: utf-8 -*-
#
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import defaultdict, deque
from typing import List, Optional, Tuple, TYPE_CHECKING, Union
from ._rules import rules
from ..._functions import lazy_product
from ...exceptions import InvalidAromaticRing


if TYPE_CHECKING:
    from chython import MoleculeContainer


class Kekule:
    __slots__ = ()

    def kekule(self: Union['Kekule', 'MoleculeContainer']) -> bool:
        """
        Convert structure to kekule form. Return True if found any aromatic ring. Set implicit hydrogen count and
        hybridization marks on atoms.

        Only one of possible double/single bonds positions will be set.
        For enumerate bonds positions use `enumerate_kekule`.
        """
        kekule = next(self.__kekule_full(), None)
        if kekule:
            bonds = self._bonds
            atoms = set()
            for n, m, b in kekule:
                bonds[n][m]._Bond__order = b
                atoms.add(n)
                atoms.add(m)
            for n in atoms:
                self._calc_implicit(n)
            self.flush_cache()
            return True
        return False

    def enumerate_kekule(self: Union['Kekule', 'MoleculeContainer']):
        """
        Enumerate all possible kekule forms of molecule.
        """
        for form in self.__kekule_full():
            copy = self.copy()
            bonds = copy._bonds
            atoms = set()
            for n, m, b in form:
                bonds[n][m]._Bond__order = b
                atoms.add(n)
                atoms.add(m)
            for n in atoms:
                copy._calc_implicit(n)
            yield copy

    def check_thiele(self, fast=True) -> bool:
        """
        Check basic aromaticity errors of molecule.

        :param fast: don't try to solve kekule form
        """
        try:
            if fast:
                self.__prepare_rings()
            else:
                next(self.__kekule_full(), None)
        except InvalidAromaticRing:
            return False
        return True

    def __fix_rings(self: 'MoleculeContainer'):
        bonds = self._bonds
        seen = set()
        for q, af, bf in rules:
            for mapping in q.get_mapping(self, automorphism_filter=False):
                match = set(mapping.values())
                if not match.isdisjoint(seen):  # prevent double patching of atoms
                    continue
                seen.update(match)

                for n, fix in af.items():
                    n = mapping[n]
                    for key, value in fix.items():
                        getattr(self, key)[n] = value
                for n, m, b in bf:
                    n = mapping[n]
                    m = mapping[m]
                    bonds[n][m]._Bond__order = b
        if seen:
            self.flush_cache()

    def __prepare_rings(self: 'MoleculeContainer'):
        atoms = self._atoms
        charges = self._charges
        radicals = self._radicals
        bonds = self._bonds
        hydrogens = self._hydrogens
        neighbors = self.neighbors

        rings = defaultdict(list)  # aromatic skeleton
        pyroles = set()

        double_bonded = defaultdict(list)
        triple_bonded = set()
        for n, m_bond in bonds.items():
            for m, bond in m_bond.items():
                bo = bond.order
                if bo == 4:
                    rings[n].append(m)
                elif bo == 2:
                    double_bonded[n].append(m)
                elif bo == 3:
                    triple_bonded.add(n)

        if not rings:
            return rings, pyroles, set()
        elif not triple_bonded.isdisjoint(rings):
            raise InvalidAromaticRing('triple bonds connected to rings')

        copy_rings = {n: ms.copy() for n, ms in rings.items()}
        for r in self.sssr:
            if set(r).issubset(rings):
                n, *_, m = r
                if n not in rings[m]:  # fix invalid structures: c1ccc-cc1
                    # remove inner ring double bonds: c1ccc=cc1
                    if n in double_bonded and m in double_bonded and m in double_bonded[n]:
                        double_bonded[n].remove(m)
                        double_bonded[m].remove(n)
                    rings[m].append(n)
                    rings[n].append(m)
                elif m in copy_rings[n]:
                    copy_rings[n].remove(m)
                    copy_rings[m].remove(n)
                for n, m in zip(r, r[1:]):
                    if n not in rings[m]:
                        if n in double_bonded and m in double_bonded and m in double_bonded[n]:
                            double_bonded[n].remove(m)
                            double_bonded[m].remove(n)
                        rings[m].append(n)
                        rings[n].append(m)
                    elif m in copy_rings[n]:
                        copy_rings[n].remove(m)
                        copy_rings[m].remove(n)

        # fix invalid smiles: c1ccccc1c2ccccc2 instead of c1ccccc1-c2ccccc2
        seen = set()
        for n, ms in copy_rings.items():
            if ms:
                seen.add(n)
                for m in ms:
                    if m not in seen:
                        rings[n].remove(m)
                        rings[m].remove(n)
                        bonds[n][m]._Bond__order = 1

        if any(len(ms) not in (2, 3) for ms in rings.values()):
            raise InvalidAromaticRing('not in ring aromatic bond or hypercondensed rings: '
                                      f'{{{", ".join(str(n) for n, ms in rings.items() if len(ms) not in (2, 3))}}}')

        # get double bonded ring atoms
        double_bonded = {n for n, ms in double_bonded.items() if ms and n in rings}
        if any(len(rings[n]) != 2 for n in double_bonded):  # double bonded never condensed
            raise InvalidAromaticRing('quinone valence error')
        for n in double_bonded:
            if atoms[n].atomic_number == 7:
                if charges[n] != 1:
                    raise InvalidAromaticRing('quinone should be charged N atom')
            elif atoms[n].atomic_number not in (6, 15, 16, 33, 34, 52) or charges[n]:
                raise InvalidAromaticRing('quinone should be neutral S, Se, Te, C, P, As atom')

        for n in rings:
            an = atoms[n].atomic_number
            ac = charges[n]
            ab = neighbors(n)
            if an == 6:  # carbon
                if ac == 0:
                    if ab not in (2, 3):
                        raise InvalidAromaticRing
                elif ac in (-1, 1):
                    if radicals[n]:
                        if ab == 2:
                            double_bonded.add(n)
                        else:
                            raise InvalidAromaticRing
                    elif ab == 3:
                        double_bonded.add(n)
                    elif ab == 2:  # benzene an[cat]ion or pyrole
                        pyroles.add(n)
                    else:
                        raise InvalidAromaticRing
                else:
                    raise InvalidAromaticRing
            elif an in (7, 15, 33):
                if ac == 0:  # pyrole or pyridine. include radical pyrole
                    if radicals[n]:
                        if ab != 2:
                            raise InvalidAromaticRing
                        double_bonded.add(n)
                    elif ab == 3:
                        if an == 7:  # pyrole only possible
                            double_bonded.add(n)
                        else:  # P(III) or P(V)H
                            pyroles.add(n)
                    elif ab == 2:
                        ah = hydrogens[n]
                        if ah is None:
                            pyroles.add(n)
                        elif ah == 1:  # only pyrole
                            double_bonded.add(n)
                        elif ah:
                            raise InvalidAromaticRing
                    elif ab != 4 or an not in (15, 33):  # P(V) in ring
                        raise InvalidAromaticRing
                elif ac == -1:  # pyrole only
                    if ab != 2 or radicals[n]:
                        raise InvalidAromaticRing
                    double_bonded.add(n)
                elif ac != 1:
                    raise InvalidAromaticRing
                elif radicals[n]:
                    if ab != 2:  # not cation-radical pyridine
                        raise InvalidAromaticRing
                elif ab == 2:  # pyrole cation or protonated pyridine
                    pyroles.add(n)
                elif ab != 3:  # not pyridine oxyde
                    raise InvalidAromaticRing
            elif an == 8:  # furan
                if ab == 2:
                    if ac == 0:
                        if radicals[n]:
                            raise InvalidAromaticRing('radical oxygen')
                        double_bonded.add(n)
                    elif ac == 1:
                        if radicals[n]:  # furan cation-radical
                            double_bonded.add(n)
                        # pyrylium
                    else:
                        raise InvalidAromaticRing('invalid oxygen charge')
                else:
                    raise InvalidAromaticRing('Triple-bonded oxygen')
            elif an in (16, 34, 52):  # thiophene
                if n not in double_bonded:  # not sulphoxyde or sulphone
                    if ab == 2:
                        if radicals[n]:
                            if ac == 1:
                                double_bonded.add(n)
                            else:
                                raise InvalidAromaticRing('S, Se, Te cation-radical expected')
                        if ac == 0:
                            double_bonded.add(n)
                        elif ac != 1:
                            raise InvalidAromaticRing('S, Se, Te cation in benzene like ring expected')
                    elif ab == 3:
                        if radicals[n]:
                            if ac:
                                raise InvalidAromaticRing('S, Se, Te ion-radical ring')
                            double_bonded.add(n)
                        elif ac == 1:
                            double_bonded.add(n)
                        elif ac:
                            raise InvalidAromaticRing('S, Se, Te invalid charge ring')
                    else:
                        raise InvalidAromaticRing('S, Se, Te hypervalent ring')
            elif an == 5:  # boron
                if ac == 0:
                    if ab == 2:
                        if radicals[n]:  # C=1O[B]OC=1
                            double_bonded.add(n)
                        else:
                            ah = hydrogens[n]
                            if ah is None:  # b1ccccc1, C=1OBOC=1 or B1C=CC=N1
                                pyroles.add(n)
                            elif ah == 1:  # C=1O[BH]OC=1 or [BH]1C=CC=N1
                                double_bonded.add(n)
                            elif ah:
                                raise InvalidAromaticRing
                    elif not radicals[n]:
                        double_bonded.add(n)
                    else:
                        raise InvalidAromaticRing
                elif ac == 1:
                    if ab == 2 and not radicals[n]:
                        double_bonded.add(n)
                    else:
                        raise InvalidAromaticRing
                elif ac == -1:
                    if ab == 2:
                        if not radicals[n]:  # C=1O[B-]OC=1 or [bH-]1ccccc1
                            pyroles.add(n)
                        # anion-radical is benzene like
                    elif radicals[n]:  # C=1O[B-*](R)OC=1
                        double_bonded.add(n)
                    else:
                        pyroles.add(n)
                else:
                    raise InvalidAromaticRing
            else:
                raise InvalidAromaticRing(f'only B, C, N, P, O, S, Se, Te possible, not: {atoms[n].atomic_symbol}')
        return rings, pyroles, double_bonded

    def __kekule_full(self):
        self.__fix_rings()  # fix bad aromatic rings
        rings, pyroles, double_bonded = self.__prepare_rings()
        atoms = set(rings)
        components = []
        while atoms:
            start = atoms.pop()
            component = {start: rings[start]}
            queue = deque([start])
            while queue:
                current = queue.popleft()
                for n in rings[current]:
                    if n not in component:
                        queue.append(n)
                        component[n] = rings[n]

            components.append(component)
            atoms.difference_update(component)

        for keks in lazy_product(*(_kekule_component(c, double_bonded & c.keys(), pyroles & c.keys())
                                   for c in components)):
            yield [x for x in keks for x in x]


def _kekule_component(rings, double_bonded, pyroles):
    # (current atom, previous atom, bond between cp atoms, path deep for cutting [None if cut impossible])
    stack: List[List[Tuple[int, int, int, Optional[int]]]]
    if double_bonded:  # start from double bonded if exists
        start = next(iter(double_bonded))
        stack = [[(next(iter(rings[start])), start, 1, 0)]]
    else:  # select not pyrole not condensed atom
        try:
            start = next(n for n, ms in rings.items() if len(ms) == 2 and n not in pyroles)
        except StopIteration:  # all pyroles. select not condensed atom.
            try:
                start = next(n for n, ms in rings.items() if len(ms) == 2)
            except StopIteration:  # fullerene?
                start = next(iter(rings))
                double_bonded.add(start)
                stack = [[(next_atom, start, 2, 0)] for next_atom in rings[start]]
            else:
                stack = [[(next_atom, start, 1, 0)] for next_atom in rings[start]]
        else:
            stack = [[(next_atom, start, 1, 0)] for next_atom in rings[start]]

    size = sum(len(x) for x in rings.values()) // 2
    path = []
    hashed_path = set()
    nether_yielded = True

    while stack:
        atom, prev_atom, bond, _ = stack[-1].pop()
        path.append((atom, prev_atom, bond))
        hashed_path.add(atom)

        if len(path) == size:
            yield path
            if nether_yielded:
                nether_yielded = False
            del stack[-1]
            if stack:
                path = path[:stack[-1][-1][-1]]
                hashed_path = {x for x, *_ in path}
        elif atom != start:
            for_stack = []
            closures = []
            loop = 0
            for next_atom in rings[atom]:
                if next_atom == prev_atom:  # only forward. behind us is the homeland
                    continue
                elif next_atom == start:
                    loop = next_atom
                elif next_atom in hashed_path:  # closure found
                    closures.append(next_atom)
                else:
                    for_stack.append(next_atom)

            if loop:  # we found starting point.
                if bond == 2:  # finish should be single bonded
                    if double_bonded:  # ok
                        stack[-1].insert(0, (loop, atom, 1, None))
                    else:
                        del stack[-1]
                        if stack:
                            path = path[:stack[-1][-1][-1]]
                            hashed_path = {x for x, *_ in path}
                        continue
                elif double_bonded:  # we in quinone ring. finish should be single bonded
                    # side-path for storing double bond or atom is quinone or pyrole
                    if for_stack or atom in double_bonded or atom in pyroles:
                        stack[-1].insert(0, (loop, atom, 1, None))
                    else:
                        del stack[-1]
                        if stack:
                            path = path[:stack[-1][-1][-1]]
                            hashed_path = {x for x, *_ in path}
                        continue
                else:  # finish should be double bonded
                    stack[-1].insert(0, (loop, atom, 2, None))
                    bond = 2  # grow should be single bonded

            if bond == 2 or atom in double_bonded:  # double in - single out. quinone has two single bonds
                for next_atom in closures:
                    path.append((next_atom, atom, 1))  # closures always single-bonded
                    stack[-1].remove((atom, next_atom, 1, None))  # remove fork from stack
                for next_atom in for_stack:
                    stack[-1].append((next_atom, atom, 1, None))
            elif len(for_stack) == 1:  # easy path grow. next bond double or include single for pyroles
                next_atom = for_stack[0]
                if next_atom in double_bonded:  # need double bond, but next atom quinone
                    if atom in pyroles:
                        stack[-1].append((next_atom, atom, 1, None))
                    else:
                        del stack[-1]
                        if stack:
                            path = path[:stack[-1][-1][-1]]
                            hashed_path = {x for x, *_ in path}
                else:
                    if atom in pyroles:  # try pyrole and pyridine
                        opposite = stack[-1].copy()
                        opposite.append((next_atom, atom, 2, None))
                        stack[-1].append((next_atom, atom, 1, len(path)))
                        stack.append(opposite)
                    else:
                        stack[-1].append((next_atom, atom, 2, None))
                        if closures:
                            next_atom = closures[0]
                            path.append((next_atom, atom, 1))  # closures always single-bonded
                            stack[-1].remove((atom, next_atom, 1, None))  # remove fork from stack
            elif for_stack:  # fork
                next_atom1, next_atom2 = for_stack
                if next_atom1 in double_bonded:  # quinone next from fork
                    if next_atom2 in double_bonded:  # bad path
                        del stack[-1]
                        if stack:
                            path = path[:stack[-1][-1][-1]]
                            hashed_path = {x for x, *_ in path}
                    else:
                        stack[-1].append((next_atom1, atom, 1, None))
                        stack[-1].append((next_atom2, atom, 2, None))
                elif next_atom2 in double_bonded:  # quinone next from fork
                    stack[-1].append((next_atom2, atom, 1, None))
                    stack[-1].append((next_atom1, atom, 2, None))
                else:  # new path
                    opposite = stack[-1].copy()
                    stack[-1].append((next_atom1, atom, 1, None))
                    stack[-1].append((next_atom2, atom, 2, len(path)))
                    opposite.append((next_atom2, atom, 1, None))
                    opposite.append((next_atom1, atom, 2, None))
                    stack.append(opposite)
            elif closures and atom not in pyroles:  # need double bond, but closure should be single bonded
                del stack[-1]
                if stack:
                    path = path[:stack[-1][-1][-1]]
                    hashed_path = {x for x, *_ in path}

    if nether_yielded:
        raise InvalidAromaticRing(f'kekule form not found for: {list(rings)}')


__all__ = ['Kekule']



# -*- coding: utf-8 -*-
#
#  Copyright 2019-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2019 Dayana Bashirova <dayana.bashirova@yandex.ru>
#  Copyright 2019 Tansu Nasyrova <tansu.nasyrova@gmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from .element import Element
from .groups import GroupXIV
from .periods import PeriodII, PeriodIII, PeriodIV, PeriodV, PeriodVI, PeriodVII


class C(Element, PeriodII, GroupXIV):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 6

    @property
    def isotopes_distribution(self):
        return {12: 0.9893, 13: 0.0107, 14: 0.0}

    @property
    def isotopes_masses(self):
        return {12: 12.0, 13: 13.003355, 14: 14.003242}

    @property
    def _common_valences(self):
        return 4,

    @property
    def _valences_exceptions(self):
        return (0, True, 3, ()), (1, False, 3, ()), (-1, False, 3, ())

    @property
    def atomic_radius(self):
        return .67


class Si(Element, PeriodIII, GroupXIV):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 14

    @property
    def isotopes_distribution(self):
        return {28: 0.922297, 29: 0.046832, 30: 0.030872}

    @property
    def isotopes_masses(self):
        return {28: 27.976927, 29: 28.976495, 30: 29.97377}

    @property
    def _common_valences(self):
        return 4,

    @property
    def _valences_exceptions(self):
        return (-2, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))),

    @property
    def atomic_radius(self):
        return 1.11


class Ge(Element, PeriodIV, GroupXIV):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 32

    @property
    def isotopes_distribution(self):
        return {70: 0.2084, 72: 0.2754, 73: 0.0773, 74: 0.3628, 76: 0.0761}

    @property
    def isotopes_masses(self):
        return {70: 69.92425, 72: 71.922076, 73: 72.923459, 74: 73.921178, 76: 75.921403}

    @property
    def _common_valences(self):
        return 4,

    @property
    def _valences_exceptions(self):
        return (-2, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))),

    @property
    def atomic_radius(self):
        return 1.25


class Sn(Element, PeriodV, GroupXIV):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 50

    @property
    def isotopes_distribution(self):
        return {112: 0.0097, 113: 0., 114: 0.0066, 115: 0.0034, 116: 0.1454, 117: 0.0768, 118: 0.2422, 119: 0.0859,
                120: 0.3258, 122: 0.0463, 124: 0.0579}

    @property
    def isotopes_masses(self):
        return {112: 111.904821, 113: 112.905171, 114: 113.902782, 115: 114.903346, 116: 115.901744, 117: 116.902954,
                118: 117.901606, 119: 118.903309, 120: 119.902197, 122: 121.903440, 124: 123.905275}

    @property
    def _common_valences(self):
        return 0, 4

    @property
    def _valences_exceptions(self):
        return ((2, False, 0, ()),
                (0, False, 0, ((2, 'O'),)), (0, False, 0, ((1, 'O'), (1, 'O'))),
                (0, False, 0, ((2, 'S'),)),

                (1, False, 0, ((1, 'C'), (1, 'C'), (1, 'C'))),
                (0, False, 1, ((1, 'C'), (1, 'C'), (1, 'C'))),

                (-2, False, 0, ((1, 'O'), (1, 'O'), (1, 'O'), (1, 'O'), (1, 'O'), (1, 'O'))))

    @property
    def atomic_radius(self):
        return 1.45


class Pb(Element, PeriodVI, GroupXIV):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 82

    @property
    def isotopes_distribution(self):
        return {204: 0.014, 206: 0.241, 207: 0.221, 208: 0.524, 210: 0.}

    @property
    def isotopes_masses(self):
        return {204: 203.973029, 206: 205.974449, 207: 206.975881, 208: 207.976636, 210: 209.984189}

    @property
    def _common_valences(self):
        return 0, 2

    @property
    def _valences_exceptions(self):
        return ((2, False, 0, ()),
                (-2, False, 0, ((1, 'O'), (1, 'O'), (1, 'O'), (1, 'O'))),
                (-2, False, 0, ((2, 'O'), (1, 'O'), (1, 'O'))),

                (0, False, 0, ((2, 'O'), (2, 'O'))),
                (0, False, 0, ((1, 'O'), (1, 'O'), (1, 'O'), (1, 'O'))),
                (0, False, 0, ((2, 'O'), (1, 'O'), (1, 'O'))),

                (0, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))),
                (0, False, 0, ((1, 'Cl'), (1, 'Cl'), (1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((1, 'C'), (1, 'C'), (1, 'C'), (1, 'C'))))

    @property
    def atomic_radius(self):
        return 1.54


class Fl(Element, PeriodVII, GroupXIV):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 114

    @property
    def isotopes_distribution(self):
        return {289: 1.0}

    @property
    def isotopes_masses(self):
        return {289: 289.190444}

    @property
    def _common_valences(self):
        return 0,

    @property
    def _valences_exceptions(self):
        return ()

    @property
    def atomic_radius(self):
        return 1.54  # unknown, taken radius of previous element in group


__all__ = ['C', 'Si', 'Ge', 'Sn', 'Pb', 'Fl']



# -*- coding: utf-8 -*-
#
#  Copyright 2018-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2021 Timur Gimadiev <timur.gimadiev@gmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CachedMethods import class_cached_property
from collections import defaultdict
from itertools import count
from typing import List, Tuple, Type, TYPE_CHECKING, Union
from ._mapping import rules as mapping_rules
from ...exceptions import MappingError


if TYPE_CHECKING:
    from chython import ReactionContainer


class StandardizeReaction:
    __slots__ = ()

    def canonicalize(self: 'ReactionContainer', fix_mapping: bool = True, *, logging=False) -> \
            Union[bool, List[Tuple[int, Tuple[int, ...], int, str]]]:
        """
        Convert molecules to canonical forms of functional groups and aromatic rings without explicit hydrogens.
        Return True if in any molecule found not canonical group.

        :param fix_mapping: Search AAM errors of functional groups.
        :param logging: return log from molecules with index of molecule.
            Otherwise return True if these groups found in any molecule.
        """
        if logging:
            total = []
        else:
            total = False
        for n, m in enumerate(self.molecules()):
            out = m.canonicalize(logging=logging)
            if out:
                if logging:
                    total.extend((n, *x) for x in out)
                else:
                    total = True

        if fix_mapping and self.fix_mapping():
            if logging:
                total.append((-1, (), -1, 'mapping fixed'))
                return total
            return True

        if total:
            self.flush_cache()
        return total

    def standardize(self: 'ReactionContainer', fix_mapping: bool = True, *, logging=False) -> \
            Union[bool, Tuple[int, Tuple[int, ...], int, str]]:
        """
        Standardize functional groups.

        :param fix_mapping: Search AAM errors of functional groups.
        :param logging: return log from molecules with index of molecule at first position.
            Otherwise return True if these groups found in any molecule.
        """
        if logging:
            total = []
        else:
            total = False
        for n, m in enumerate(self.molecules()):
            out = m.standardize(logging=logging)
            if out:
                if logging:
                    total.extend((n, *x) for x in out)
                else:
                    total = True

        if fix_mapping and self.fix_mapping():
            if logging:
                total.append((-1, (), -1, 'mapping fixed'))
                return total
            return True

        if total:
            self.flush_cache()
        return total

    def standardize_charges(self: 'ReactionContainer', *, logging=False) -> \
            Union[bool, List[Tuple[int, Tuple[int, ...]]]]:
        """
        Set canonical positions of charges in heterocycles and some nitrogen compounds.

        :param logging: return list of changed molecules with tuple of changed atoms.
        """
        if logging:
            total = []
        else:
            total = False
        for n, m in enumerate(self.molecules()):
            out = m.standardize_charges(logging=logging)
            if out:
                if logging:
                    total.append((n, tuple(out)))
                else:
                    total = True
        if total:
            self.flush_cache()
        return total

    def thiele(self: 'ReactionContainer') -> bool:
        """
        Convert structures to aromatic form.
        Return True if in any molecule found kekule ring
        """
        total = False
        for m in self.molecules():
            if m.thiele() and not total:
                total = True
        if total:
            self.flush_cache()
        return total

    def kekule(self: 'ReactionContainer') -> bool:
        """
        Convert structures to kekule form. Works only for Molecules.
        Return True if in any molecule found aromatic ring
        """
        total = False
        for m in self.molecules():
            if m.kekule() and not total:
                total = True
        if total:
            self.flush_cache()
        return total

    def clean_isotopes(self: 'ReactionContainer') -> bool:
        """
        Clean isotope marks for all molecules in reaction.
        Returns True if in any molecule found isotope.
        """
        flag = False
        for m in self.molecules():
            if m.clean_isotopes() and not flag:
                flag = True
        if flag:
            self.flush_cache()
        return flag

    def clean_stereo(self: 'ReactionContainer'):
        """
        Remove stereo data
        """
        for m in self.molecules():
            m.clean_stereo()
        self.flush_cache()

    def check_valence(self: 'ReactionContainer') -> List[Tuple[int, Tuple[int, ...]]]:
        """
        Check valences of all atoms of all molecules.

        Works only on molecules with aromatic rings in Kekule form.
        :return: list of invalid molecules with invalid atoms lists
        """
        out = []
        for n, m in enumerate(self.molecules()):
            c = m.check_valence()
            if c:
                out.append((n, tuple(c)))
        return out

    def fix_mapping(self: Union['ReactionContainer', 'StandardizeReaction'], *, logging: bool = False) -> \
            Union[bool, List[Union[Tuple[str, List[int]], Tuple[int, str, str, Tuple[int, ...]]]]]:
        """
        Fix atom-to-atom mapping of some functional groups. Return True if found AAM errors.
        """
        log = []
        seen = set()
        if not self:
            return False

        for r_pattern, p_pattern, fix in mapping_rules:
            found = []
            for m in self.reactants:
                for mapping in r_pattern.get_mapping(m, automorphism_filter=False):
                    if mapping[1] not in seen:
                        found.append(({fix.get(k, k): v for k, v in mapping.items()},
                                      {mapping[k]: mapping[v] for k, v in fix.items()}))

            if not found:
                continue
            for m in self.products:
                for mapping in p_pattern.get_mapping(m, automorphism_filter=False):
                    atom = mapping[1]
                    if atom in seen:
                        continue
                    for n, (k, v) in enumerate(found):
                        if k == mapping:
                            break
                    else:
                        continue

                    del found[n]
                    m.remap(v)
                    seen.add(atom)
                    log.append(('group remap', list(v)))
        if seen:
            self.flush_cache()
            flag = True
            seen = set()
        else:
            flag = False

        for rule_num, (bad_query, good_query, fix, valid) in enumerate(self.__remapping_compiled_rules):
            cgr = ~self
            first_free = max(cgr) + 1
            free_number = count(first_free)
            cgr_c = set(cgr.center_atoms)
            del self.__dict__['__cached_method_compose']

            flag_m = False
            for mapping in bad_query.get_mapping(cgr, automorphism_filter=False):
                if not seen.isdisjoint(mapping.values()):  # prevent matching same RC
                    continue
                mapping = {mapping[n]: next(free_number) if m is None else mapping[m] for n, m in fix.items()}
                flag_m = True
                reverse = {m: n for n, m in mapping.items()}
                for m in self.products:
                    m.remap(mapping)

                check = ~self
                check_c = set(check.center_atoms)
                delta = check_c - cgr_c

                for m in good_query.get_mapping(check, automorphism_filter=False):
                    if valid.issubset(m) and delta.issubset(m.values()):
                        seen.update(mapping)
                        log.append((rule_num, str(bad_query), str(good_query), tuple(mapping.values())))
                        flag = True
                        break
                else:
                    # restore old mapping
                    for m in self.products:
                        m.remap(reverse)
                    del self.__dict__['__cached_method_compose']
                    free_number = count(first_free)
                    continue
                break
            else:
                if flag_m:
                    log.append((rule_num, str(bad_query), str(good_query), ()))
        if seen:
            self.flush_cache()
        if logging:
            return log
        return flag

    @classmethod
    def load_remapping_rules(cls: Type['ReactionContainer'], reactions):
        """
        Load AAM fixing rules. Required pairs of bad mapped and good mapped reactions.
        Reactants in pairs should be fully equal (equal molecules and equal atom orders).
        Products should be equal but with different atom numbers.
        """
        rules = []
        for bad, good in reactions:
            if str(bad) != str(good):
                raise ValueError('bad and good reaction should be equal')

            cgr_good, cgr_bad = ~good, ~bad
            gc = cgr_good.substructure({y for x in cgr_good.center_atoms for y in cgr_good._bonds[x]} |
                                       set(cgr_good.center_atoms))
            bc = cgr_bad.substructure({y for x in cgr_bad.center_atoms for y in cgr_bad._bonds[x]} |
                                      set(cgr_bad.center_atoms))
            atoms = set(bc) | set(gc)

            pr_g, pr_b, re_g, re_b = set(), set(), set(), set()
            for pr in good.products:
                pr_g.update(pr)
            for pr in bad.products:
                pr_b.update(pr)
            for pr in good.reactants:
                re_g.update(pr)
            for pr in bad.reactants:
                re_b.update(pr)
            atoms.update((re_b.difference(pr_b)).intersection(pr_g))

            strange_atoms = pr_b.difference(pr_g)
            atoms.update(strange_atoms)

            bad_query = cgr_bad.substructure(atoms.intersection(cgr_bad))
            good_query = cgr_good.substructure(atoms.intersection(cgr_good))

            fix = {}
            for mb, mg in zip(bad.products, good.products):
                fix.update({k: v for k, v in zip(mb, mg) if k != v and k in atoms})

            valid = set(fix).difference(strange_atoms)
            rules.append((bad_query, good_query, fix, valid))

        cls.__class_cache__[cls] = {'_StandardizeReaction__remapping_compiled_rules': tuple(rules)}

    @class_cached_property
    def __remapping_compiled_rules(self):
        return ()

    def implicify_hydrogens(self: 'ReactionContainer') -> int:
        """
        Remove explicit hydrogens if possible

        :return: number of removed hydrogens
        """
        total = 0
        for m in self.molecules():
            total += m.implicify_hydrogens()
        if total:
            self.flush_cache()
        return total

    def explicify_hydrogens(self: 'ReactionContainer') -> int:
        """
        Add explicit hydrogens to atoms

        :return: number of added atoms
        """
        total = 0
        start_map = 0
        for m in self.molecules():
            map_ = max(m, default=0)
            if map_ > start_map:
                start_map = map_

        mapping = defaultdict(list)
        for m in self.reactants:
            maps = m.explicify_hydrogens(return_maps=True, start_map=start_map + 1)
            if maps:
                for n, h in maps:
                    mapping[n].append(h)
                start_map = maps[-1][1]
                total += len(maps)

        for m in self.reagents:
            maps = m.explicify_hydrogens(return_maps=True, start_map=start_map + 1)
            if maps:
                start_map = maps[-1][1]
                total += len(maps)

        for m in self.products:
            maps = m.explicify_hydrogens(return_maps=True, start_map=start_map + 1)
            if maps:
                total += len(maps)
                remap = {}
                free = []
                for n, h in maps:
                    if n in mapping and mapping[n]:
                        remap[h] = mapping[n].pop()
                        free.append(h)
                    elif free:
                        remap[h] = start_map = free.pop(0)
                    else:
                        start_map = h
                m.remap(remap)

        if total:
            self.flush_cache()
        return total

    def remove_reagents(self: 'ReactionContainer', *, keep_reagents: bool = False) -> bool:
        """
        Preprocess reaction according to mapping, using the following idea: molecules(each separated graph) will be
        placed to reagents if it is not changed in the reaction (no bonds, charges reorders)

        Return True if any reagent found.
        """
        cgr = ~self
        if cgr.center_atoms:
            active = set(cgr.center_atoms)
            reactants = []
            products = []
            reagents = set(self.reagents)
            for i in self.reactants:
                if not active.isdisjoint(i):
                    reactants.append(i)
                else:
                    reagents.add(i)
            for i in self.products:
                if not active.isdisjoint(i):
                    products.append(i)
                else:
                    reagents.add(i)
            if keep_reagents:
                tmp = []
                for m in self.reagents:
                    if m in reagents:
                        tmp.append(m)
                        reagents.discard(m)
                tmp.extend(reagents)
                reagents = tuple(tmp)
            else:
                reagents = ()

            if len(reactants) != len(self.reactants) or len(products) != len(self.products) or \
                    len(reagents) != len(self.reagents):
                self._ReactionContainer__reactants = tuple(reactants)
                self._ReactionContainer__products = tuple(products)
                self._ReactionContainer__reagents = reagents
                self.flush_cache()
                self.fix_positions()
                return True
            return False
        raise MappingError("Reaction center is absent according to mapping")

    def contract_ions(self: 'ReactionContainer') -> bool:
        """
        Contract ions into salts (Molecules with disconnected components).
        Note: works only for unambiguous cases. e.g. equal anions/cations and different or equal cations/anions.

        Return True if any ions contracted.
        """
        neutral, cations, anions, total = _sift_ions(self.reagents)
        salts = _contract_ions(anions, cations, total)
        if salts:
            neutral.extend(salts)
            self._ReactionContainer__reagents = tuple(neutral)
            changed = True
        else:
            changed = False

        neutral, cations, anions, total = _sift_ions(self.reactants)
        salts = _contract_ions(anions, cations, total)
        if salts:
            anions_order = {frozenset(m): n for n, m in enumerate(anions)}
            cations_order = {frozenset(m): n for n, m in enumerate(cations)}
            neutral.extend(salts)
            self._ReactionContainer__reactants = tuple(neutral)
            changed = True
        else:
            anions_order = cations_order = {}

        neutral, cations, anions, total = _sift_ions(self.products)
        if cations and anions:
            anions.sort(key=lambda x: anions_order.get(frozenset(x), -1))
            cations.sort(key=lambda x: cations_order.get(frozenset(x), -1))
        salts = _contract_ions(anions, cations, total)
        if salts:
            neutral.extend(salts)
            self._ReactionContainer__products = tuple(neutral)
            changed = True

        if changed:
            self.flush_cache()
            self.fix_positions()
            return True
        return False


def _sift_ions(mols):
    anions = []
    cations = []
    neutral = []
    total = 0
    for m in mols:
        c = int(m)
        total += c
        if c > 0:
            cations.append(m)
        elif c < 0:
            anions.append(m)
        else:
            neutral.append(m)
    return neutral, cations, anions, total


def _contract_ions(anions, cations, total):
    salts = []
    if not anions or not cations:  # nothing to contract
        return
    # check ambiguous cases
    if total > 0:
        if len(cations) > 1:  # deficit of anions
            return
    elif total < 0:
        if len(anions) > 1:  # deficit of cations
            return
    elif len(set(anions)) > 1 and len(set(cations)) > 1:  # different anions and cations
        return

    anions = anions.copy()
    cations = cations.copy()
    while anions:
        ct = cations.pop()
        an = anions.pop()
        shift_x = ct._fix_plane_mean(0) + 1
        shift_x = an._fix_plane_mean(shift_x)
        salt = ct | an
        while True:
            c = int(salt)
            if c > 0:
                an = anions.pop()
                shift_x = an._fix_plane_mean(shift_x) + 1
                salt = salt | an
            elif c < 0:
                ct = cations.pop()
                shift_x = ct._fix_plane_mean(shift_x) + 1
                salt = salt | ct
            else:
                break
        salts.append(salt)
    return salts


__all__ = ['StandardizeReaction']
